import L from 'leaflet';
import ready from 'document-ready';

/**
 * ------------------------------------------------------------------------
 * Constants
 * ------------------------------------------------------------------------
 */

const DATA_KEY = 'map';

const Default = {
  markerCluster: true,
  tileProvider: 'OSMFrance'
};

const Selector = {
  MAP_CONTAINER: `[data-${DATA_KEY}]`
};

const TileProvider = {
  OSM: {
    url: 'https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
    options: {
      maxZoom: 19,
      attribution: '&copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap</a> contributors'
    }
  },
  OSMFrance: {
    url: '//{s}.tile.openstreetmap.fr/osmfr/{z}/{x}/{y}.png',
    options: {
      maxZoom: 20,
      attribution: '&copy; Openstreetmap France | &copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap</a> contributors'
    }
  }
};

/**
 * ------------------------------------------------------------------------
 * Function Helpers
 * ------------------------------------------------------------------------
 */

function parseJSONFromElement(element) {
  return element ? JSON.parse(element.textContent) : null;
}

/**
 * ------------------------------------------------------------------------
 * Class Definition
 * ------------------------------------------------------------------------
 */

/* global layer_icon_url */
/* global layer_shadow_url */

class Map {
  constructor(element, config) {
    this._element = element;
    this._config = this._getConfig(config);

    this._map = this._getMap();
    this._icon = new L.Icon({
      iconUrl: layer_icon_url,
      shadowUrl: layer_shadow_url
    });

    if (this._config.markers) {
      const markers = this._config.markers.reduce((result, item) => {
        if (item.lat && item.lng) {
          result.push(this._createMarker(item));
        }
        return result;
      }, []);

      if (this._config.markerCluster) {
        this._map.addLayer(this._createMarkerClusterGroup(markers));
      } else {
        markers.forEach((marker) => marker.addTo(this._map));
      }
    }
  }

  // Getters

  static get Default() {
    return Default;
  }

  // Private

  _getConfig(config) {
    config = {
      ...Default,
      ...config
    };

    if (!(config.tileProvider in TileProvider)) {
      throw new Error(`Unknown tile provider name '${config.tileProvider}'`);
    }

    if (!config.mapOptions) {
      throw new Error('No map options have been specified');
    }

    if (config.markerCluster && !L.markerClusterGroup) {
      config.markerCluster = false;
    }

    return config;
  }

  _getMap() {
    const map = L.map(this._element, this._config.mapOptions);

    const provider = TileProvider[this._config.tileProvider];
    L.tileLayer(provider.url, provider.options).addTo(map);

    return map;
  }

  _createMarker(options) {
    const marker = L.marker([options.lat, options.lng], { icon: this._icon });

    if (options.popup) {
      marker.bindPopup(options.popup);
    }

    return marker;
  }

  _createMarkerClusterGroup(markers = []) {
    const group = L.markerClusterGroup();

    markers.forEach((marker) => group.addLayer(marker));

    return group;
  }
}

/**
 * ------------------------------------------------------------------------
 * Data Api implementation
 * ------------------------------------------------------------------------
 */

ready(() => {
  /**
   * Construis une carte dans les éléments du document ayant un attribut
   * `data-map`. Cet attribut doit avoir comme valeur l'id d'un autre élément
   * de type <script>, contenant les options au format JSON à passer à Map().
   */
  Array.from(document.querySelectorAll(Selector.MAP_CONTAINER))
    .forEach((container) => {
      const optionsElementId = container.getAttribute(`data-${DATA_KEY}`);
      const options = optionsElementId &&
        parseJSONFromElement(document.getElementById(optionsElementId));

      return new Map(container, options);
    });
});

export default Map;
