from collections import OrderedDict

from django import forms
from django.contrib.auth import forms as auth_forms
from django.db import transaction
from django.urls import reverse

from betterforms.multiform import MultiModelForm
from registration import forms as registration_forms
from registration.models import RegistrationProfile
from tapeforms.contrib.foundation import FoundationTapeformMixin
from wagtail.users.forms import UserCreationForm as WagtailUserCreationForm
from wagtail.users.forms import UserEditForm as WagtailUserEditForm

from boiteaasso.base.forms import AdresseFormMixin

from .models import Association, Personne, UserAccount


class AuthFormMixin(FoundationTapeformMixin):
    """
    Mixin pour les formulaires liés à l'authentification.
    """

    field_template = 'tapeforms/fields/auth.html'
    widget_css_class = 'input-group-field'

    def apply_widget_options(self, field_name):
        field = self.fields[field_name]
        field.widget.attrs['placeholder'] = field.label


class AuthenticationForm(AuthFormMixin, auth_forms.AuthenticationForm):
    """
    Formulaire d'authentification d'un compte utilisateur.
    """

    error_messages = {
        'invalid_login': (
            "L'adresse mail et le mot de passe ne correspondent pas, "
            "veuillez réessayer."
        ),
        'inactive': (
            "Le compte est désactivé. Veuillez nous contacter pour plus "
            "d'informations."
        ),
        'not_activated': (
            "Le compte n'a pas encore été activé, vous avez dû recevoir un "
            "courriel contenant un lien d'activation. Si vous ne l'avez plus, "
            "vous pouvez <a href=\"%(resend_url)s\">demander à le "
            "renvoyer</a>."
        ),
    }

    def get_invalid_login_error(self):
        """Adapte l'erreur si le compte est en attente d'activation."""
        try:
            RegistrationProfile.objects.get(
                user__email=self.cleaned_data.get('username'), activated=False
            )
        except RegistrationProfile.DoesNotExist:
            return forms.ValidationError(
                self.error_messages['invalid_login'], code='invalid_login'
            )
        else:
            return forms.ValidationError(
                self.error_messages['not_activated'],
                code='not_activated',
                params={
                    'resend_url': reverse(
                        'accounts:registration_resend_activation'
                    )
                },
            )


class PasswordResetForm(AuthFormMixin, auth_forms.PasswordResetForm):
    """
    Formulaire de réinitialisation du mot de passe.
    """

    email = forms.EmailField(label="Adresse mail", max_length=254)


class SetPasswordForm(FoundationTapeformMixin, auth_forms.SetPasswordForm):
    """
    Formulaire de changement de mot de passe sans l'ancien.
    """

    def apply_widget_options(self, field_name):
        field = self.fields[field_name]
        field.widget.attrs['autocomplete'] = 'new-password'


class ResendActivationForm(
    AuthFormMixin, registration_forms.ResendActivationForm
):
    """
    Formulaire de renvoi du courriel d'activation d'un compte.
    """

    email = forms.EmailField(label="Adresse mail", max_length=254)


class AccountCreationForm(FoundationTapeformMixin, auth_forms.UserCreationForm):
    """
    Formulaire de création d'un compte utilisateur de base.
    """

    class Meta:
        model = UserAccount
        fields = ('email',)


class AssociationForm(
    FoundationTapeformMixin, AdresseFormMixin, forms.ModelForm
):
    """
    Formulaire d'édition d'une association.
    """

    responsable_tel = forms.CharField(
        label="Téléphone", min_length=10, max_length=15
    )

    class Meta:
        model = Association
        fields = '__all__'
        labels = {
            'responsable_nom': "Nom",
            'responsable_prenom': "Prénom",
            'responsable_fonction': "Fonction dans l'association",
            'responsable_email': "Adresse mail",
        }


class AssociationRegistrationForm(MultiModelForm):
    """
    Formulaire de création d'un compte pour une association.
    """

    form_classes = OrderedDict(
        (('account', AccountCreationForm), ('association', AssociationForm))
    )

    @transaction.atomic
    def save(self, *args, **kwargs):
        association_form = self.forms['association']
        # instancie l'association et lui définis le nouveau compte
        association = association_form.save(commit=False)
        association.account = self.forms['account'].save()
        # enregistre l'association et les éventuelles relations
        association.save()
        association_form.save_m2m()
        # retourne le compte, nécessaire pour RegistrationView
        return association.account


class PersonneForm(FoundationTapeformMixin, AdresseFormMixin, forms.ModelForm):
    """
    Formulaire d'édition d'une personne.
    """

    tel = forms.CharField(
        label="Téléphone", min_length=10, max_length=15, required=False
    )

    class Meta:
        model = Personne
        fields = '__all__'


class PersonneRegistrationForm(MultiModelForm):
    """
    Formulaire de création d'un compte pour une personne.
    """

    form_classes = OrderedDict(
        (('account', AccountCreationForm), ('personne', PersonneForm))
    )

    @transaction.atomic
    def save(self, *args, **kwargs):
        personne_form = self.forms['personne']
        # instancie la personne et lui définis le nouveau compte
        personne = personne_form.save(commit=False)
        personne.account = self.forms['account'].save()
        # enregistre la personne et les éventuelles relations
        personne.save()
        personne_form.save_m2m()
        # retourne le compte, nécessaire pour RegistrationView
        return personne.account


# FORMULAIRES POUR WAGTAIL
# ------------------------------------------------------------------------------
# https://docs.wagtail.io/en/stable/advanced_topics/customisation/custom_user_models.html

standard_fields = set(['email', 'is_superuser', 'groups'])


class WagtailAccountFormMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # supprime les champs non implémentés ici
        for field in ('first_name', 'last_name'):
            if field in self.fields:
                del self.fields[field]


class WagtailAccountCreationForm(
    WagtailAccountFormMixin, WagtailUserCreationForm
):
    class Meta(WagtailUserCreationForm.Meta):
        fields = standard_fields


class WagtailAccountEditForm(WagtailAccountFormMixin, WagtailUserEditForm):
    class Meta(WagtailUserCreationForm.Meta):
        fields = set(['is_active']) | standard_fields
