from collections import namedtuple

from django.contrib.auth.models import (
    AbstractBaseUser,
    BaseUserManager,
    Group,
    PermissionsMixin,
)
from django.core.exceptions import ValidationError
from django.core.mail import send_mail
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone

from djchoices import ChoiceItem, DjangoChoices
from wagtail.admin.panels import FieldPanel, MultiFieldPanel

from boiteaasso.base.models import AdresseMixin, OptionalAdresseMixin

"""Liste des classes de types d'utilisateurs."""
USER_TYPE_CLASSES = set()

"""Informations sur un type d'utilisateur."""
UserTypeInfo = namedtuple(
    'UserTypeInfo', ['name', 'verbose_name', 'verbose_name_plural']
)


class UserAccountManager(BaseUserManager):
    use_in_migrations = True

    def create_user(self, email, password=None, **extra_fields):
        """
        Crée et enregistre un compte utilisateur avec l'adresse mail et le
        mot de passe donnés.
        """
        extra_fields.setdefault('is_superuser', False)
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, email, password, **extra_fields):
        """
        Crée et enregistre un compte super-utilisateur.
        """
        extra_fields.setdefault('is_superuser', True)
        if extra_fields.get('is_superuser') is not True:
            raise ValueError(
                "Un compte super-utilisateur doit avoir is_superuser=True."
            )
        return self.create_user(email, password, **extra_fields)

    def admins(self):
        """
        Filtre avec les administrateurs actifs uniquement.
        """
        return self.filter(is_active=True, is_superuser=True)


class UserAccount(AbstractBaseUser, PermissionsMixin):
    """
    Un compte utilisateur au sein de l'application.
    """

    email = models.EmailField(
        "adresse mail",
        unique=True,
        error_messages={
            'unique': "Un compte utilisateur existe déjà avec cette adresse."
        },
    )
    is_active = models.BooleanField(
        "actif",
        default=True,
        help_text=(
            "Précise si le compte utilisateur doit être considéré comme "
            "actif. Vous pouvez le désactiver plutôt que de le supprimer."
        ),
    )
    date_joined = models.DateTimeField(
        "date d'inscription", default=timezone.now, editable=False
    )
    # Champs supplémentaires pour l'application
    full_name = models.CharField(max_length=150, blank=True, editable=False)
    short_name = models.CharField(max_length=150, blank=True, editable=False)

    objects = UserAccountManager()

    EMAIL_FIELD = 'email'
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    class Meta:
        verbose_name = "compte utilisateur"
        verbose_name_plural = "comptes utilisateurs"
        ordering = ['full_name', 'email']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # récupère le type d'utilisateur actuel
        self.__details = None
        self.refresh_user_type()

    @property
    def is_staff(self):
        # propriété non utilisée dans l'application, à part par l'admin
        # de Django, on se base donc sur la valeur de `is_superuser`
        return self.is_superuser

    def clean(self):
        super().clean()
        self.email = self.__class__.objects.normalize_email(self.email)

    def get_full_name(self):
        return self.full_name

    def get_short_name(self):
        return self.short_name

    def email_user(self, subject, message, from_email=None, **kwargs):
        """Envoi un courriel à ce compte utilisateur."""
        send_mail(subject, message, from_email, [self.email], **kwargs)

    # Accès au type d'utilisateur et ses détails -----------------------------

    @property
    def details(self):
        """Retourne les détails de l'utilisateur pour son type."""
        return self.__details

    @property
    def type_info(self):
        """Retourne les informations sur le type de cet utilisateur."""
        return self.details.type_info if self.details is not None else None

    def refresh_user_type(self, details=None):
        """Détermine et stocke le type actuel de cet utilisateur."""
        if details is not None:
            # vérifie l'instance donnée avant de la stocker
            if not isinstance(details, AbstractUserType):
                raise ValueError(
                    "'details' doit être une instance de AbstractUserType."
                )
            if details.account_id != self.id:
                raise ValueError(
                    "{} ne pointe pas vers ce compte utilisateur.".format(
                        details
                    )
                )
        elif self.pk is not None:
            user_types = set()
            # cherche toutes les relations définies pour ce compte
            for model in USER_TYPE_CLASSES:
                details = getattr(self, model.type_info.name, None)
                if details is not None and details.pk is not None:
                    user_types.add(details)
            if not user_types:
                details = None
            elif len(user_types) == 1:
                details = user_types.pop()
            else:
                raise ValueError(
                    "{} a plusieurs types de définis : {}".format(
                        self, user_types
                    )
                )
        self.__details = details

    def add_to_group_for_type(self):
        """Ajoute l'utilisateur dans le groupe de son type, s'il existe."""
        if not self.type_info:
            return

        name = self.type_info.verbose_name
        name_plural = '{}s'.format(name)

        # cherche et ajoute l'utilisateur dans les groupes qui ont comme nom
        # celui de son type, au singulier ou au pluriel
        self.groups.add(
            *Group.objects.filter(
                Q(name__iexact=name) | Q(name__iexact=name_plural)
            )
        )

    def get_admin_details_url(self):
        if self.type_info:
            return reverse(
                'accounts_{}_modeladmin_inspect'.format(self.type_info.name),
                args=[self.details.pk],
            )
        return reverse('wagtailusers_users:edit', args=[self.pk])


# CLASSES DE BASE POUR LES TYPES D'UTILISATEURS
# ------------------------------------------------------------------------------


class BaseUserTypeManager(models.Manager):
    def create(self, **kwargs):
        # crée s'il le faut en plus le compte pour l'utilisateur
        if 'account' not in kwargs and 'email' in kwargs:
            kwargs['account'] = UserAccount.objects.create_user(
                **{
                    field: kwargs.pop(field)
                    for field in ('email', 'password', 'is_active')
                    if field in kwargs
                }
            )
        return super().create(**kwargs)


class UserTypeBase(models.base.ModelBase):
    """Métaclasse des types d'utilisateurs."""

    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        # traite seulement les classes non abstraites
        if not new_class._meta.abstract:
            if hasattr(new_class, 'type_info'):
                raise AttributeError("L'attribut 'type_info' est réservé.")
            new_class.type_info = UserTypeInfo(
                name=new_class._meta.model_name,
                verbose_name=new_class._meta.verbose_name,
                verbose_name_plural=new_class._meta.verbose_name_plural,
            )
            USER_TYPE_CLASSES.add(new_class)
        return new_class


class AbstractUserType(models.Model, metaclass=UserTypeBase):
    """Définis la base commune d'un type d'utilisateur.

    Les classes définissant un type d'utilisateur et ses détails doivent
    hériter de cette classe abstraite. Elles seront automatiquement ajoutées
    aux types d'utilisateurs disponibles, contenus dans `USER_TYPE_CLASSES`.
    La propriété ``type_info`` sera ajoutée automatiquement pour la classe,
    et retournera un objet de type `UserTypeInfo`.

    Une relation ``account`` de type ``OneToOneField`` est créé afin de lier
    les détails d'un utilisateur à son compte. Ce champ est requis et est
    unique pour le modèle, afin d'empêcher qu'un compte ait plusieurs
    types d'associés. Cette vérification est aussi renforcée lors de
    l'enregistrement et la validation.

    Enfin, plusieurs méthodes sont à implémenter par les sous-classes, comme
    les noms complet et court à utiliser pour l'utilisateur.
    """

    account = models.OneToOneField(
        UserAccount,
        on_delete=models.CASCADE,
        related_name='%(class)s',
        related_query_name='%(class)ss',
        editable=False,
        verbose_name="compte utilisateur",
    )

    objects = BaseUserTypeManager()

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # stocke l'id du compte utilisateur pour une entrée existante
        self.__account_id = self.account_id if self.pk is not None else None

    def __str__(self):
        return self.get_full_name()

    def clean(self):
        # le compte peut ne pas exister à la vérification
        if hasattr(self, 'account'):
            self._validate_account()
        super().clean()

    def save(self, *args, **kwargs):
        self._validate_account(True)
        super().save(*args, **kwargs)

        # mets à jour l'id du compte utilisateur actuel
        self.__account_id = self.account_id
        # TODO: déplacer ces opérations dans les signaux ?
        # mets à jour le nom complet et court du compte utilisateur
        self.account.full_name = self.get_full_name()
        self.account.short_name = self.get_short_name()
        self.account.save()
        # mets à jour le type d'utilisateur pour le compte
        self.account.refresh_user_type(self)

    def _validate_account(self, is_saving=False):
        """Vérifie le compte utilisateur pour clean() et save().

        Renforce la vérification du compte utilisateur à la validation
        (via `clean()`) et à l'enregistrement (via `save()`) afin d'éviter
        que plusieurs types d'utilisateurs soient associés à un compte. Il
        est nécessaire d'appeler cette méthode dans les deux cas vu qu'il
        est possible d'enregistrer l'objet sans le valider.
        """
        error = None

        if self.__account_id is None:
            if (
                self.account.details is not None
                and self.account.details.pk != self.pk
            ):
                error = "Un type d'utilisateur est déjà associé au compte."
        elif self.__account_id != self.account_id:
            error = "Le compte utilisateur ne peut pas être modifié."

        if error:
            if is_saving:
                raise ValueError(error)
            raise ValidationError({'account': error})

    # Méthodes à implémenter --------------------------------------------------

    def get_full_name(self):
        """Retourne le nom complet de ce compte utilisateur."""
        raise NotImplementedError(
            "Les sous-classes doivent implémenter get_full_name()."
        )

    def get_short_name(self):
        """Retourne le nom court de ce compte utilisateur."""
        raise NotImplementedError(
            "Les sous-classes doivent implémenter get_short_name()."
        )


# DEFINITION DES TYPES D'UTILISATEURS
# ------------------------------------------------------------------------------


class AssociationActiviteType(DjangoChoices):
    """
    Les types d'activités pour une association.
    """

    culture = ChoiceItem('culture', "Culture")
    loisirs = ChoiceItem('loisirs', "Loisirs")
    sport = ChoiceItem('sport', "Sport")
    autre = ChoiceItem('autre', "Autre")


class Association(AbstractUserType, AdresseMixin):
    """
    Détails d'une association.
    """

    nom = models.CharField(
        "nom",
        max_length=150,
        unique=True,
        error_messages={'unique': "Une association existe déjà avec ce nom."},
    )
    sigle = models.CharField("sigle", max_length=30, blank=True)
    activite_type = models.CharField(
        "type d'activité",
        max_length=10,
        choices=AssociationActiviteType.choices,
    )
    # Responsable de l'association
    responsable_nom = models.CharField("nom du responsable", max_length=150)
    responsable_prenom = models.CharField(
        "prénom du responsable", max_length=30
    )
    responsable_fonction = models.CharField(
        "fonction du responsable", max_length=100
    )
    responsable_email = models.EmailField("adresse mail du responsable")
    responsable_tel = models.CharField(
        "téléphone du responsable", max_length=15
    )

    class Meta:
        verbose_name = "association"
        verbose_name_plural = "associations"

    def get_full_name(self):
        return self.nom

    def get_short_name(self):
        return self.sigle or self.nom

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        MultiFieldPanel(
            [
                FieldPanel('nom'),
                FieldPanel('sigle'),
                FieldPanel('activite_type'),
            ],
            heading="Informations",
        ),
        MultiFieldPanel(AdresseMixin.adresse_panels, heading="Adresse"),
        MultiFieldPanel(
            [
                FieldPanel('responsable_nom'),
                FieldPanel('responsable_prenom'),
                FieldPanel('responsable_fonction'),
                FieldPanel('responsable_email'),
                FieldPanel('responsable_tel'),
            ],
            heading="Responsable",
        ),
    ]


class Personne(AbstractUserType, OptionalAdresseMixin):
    """
    Détails d'une personne physique.
    """

    prenom = models.CharField("prénom ou pseudo", max_length=50)
    nom = models.CharField("nom", max_length=50, null=True, blank=True)
    tel = models.CharField("téléphone", max_length=15, null=True, blank=True)

    class Meta:
        verbose_name = "compte personne physique"
        verbose_name_plural = "comptes personnes physiques"

    def get_full_name(self):
        return "{} {}".format(self.prenom, self.nom)

    def get_short_name(self):
        return self.prenom

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        MultiFieldPanel(
            [FieldPanel('prenom'), FieldPanel('nom'), FieldPanel('tel')],
            heading="Informations",
        ),
        MultiFieldPanel(AdresseMixin.adresse_panels, heading="Adresse"),
    ]
