from django.contrib.auth.models import Permission

import factory
from factory import Faker
from factory.django import DjangoModelFactory

from ..models import Association, AssociationActiviteType, Personne, UserAccount
from .models import Organization, Person

Faker._DEFAULT_LOCALE = 'fr_FR'


class UserAccountFactory(DjangoModelFactory):
    email = Faker('email')

    class Meta:
        model = UserAccount

    @factory.post_generation
    def password(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            self.set_password(extracted)

    @factory.post_generation
    def permissions(self, create, extracted, codenames=None, **kwargs):
        if not create:
            return
        if codenames:
            self.user_permissions.set(
                Permission.objects.filter(codename__in=codenames)
            )
        if extracted:
            for perm in extracted:
                self.user_permissions.add(perm)


class UserTypeFactory(DjangoModelFactory):
    account = factory.SubFactory(UserAccountFactory)

    @classmethod
    def build(cls, email=None, **kwargs):
        obj = super().build(**kwargs)
        if email and not hasattr(obj, 'account'):
            obj.account = UserAccountFactory.build(email=email)
        return obj


# MODÈLES DE TEST
# ---------------


class OrganizationFactory(UserTypeFactory):
    name = Faker('company')

    class Meta:
        model = Organization


class PersonFactory(UserTypeFactory):
    first_name = Faker('first_name')
    last_name = Faker('last_name')

    class Meta:
        model = Person


# MODÈLES RÉELS
# -------------

ASSOCIATION_PERMISSIONS = [
    'add_annonce',
    'contact_annonce',
    'add_offreemploi',
    'contact_offreemploi',
    'add_formation',
    'subscribe_formation',
]


class AssociationFactory(UserTypeFactory):
    account = factory.SubFactory(
        UserAccountFactory, permissions__codenames=ASSOCIATION_PERMISSIONS
    )

    nom = Faker('company')
    activite_type = Faker(
        'random_element', elements=AssociationActiviteType.values.keys()
    )

    responsable_nom = Faker('last_name')
    responsable_prenom = Faker('first_name')
    responsable_fonction = Faker('job')
    responsable_email = Faker('email')
    responsable_tel = Faker('phone_number')

    class Meta:
        model = Association


class AssociationUserAccountFactory(UserAccountFactory):
    permissions__codenames = ASSOCIATION_PERMISSIONS
    association = factory.RelatedFactory(AssociationFactory, 'account')


PERSONNE_PERMISSIONS = [
    'contact_annonce',
    'contact_offreemploi',
    'subscribe_formation',
]


class PersonneFactory(UserTypeFactory):
    account = factory.SubFactory(
        UserAccountFactory, permissions__codenames=PERSONNE_PERMISSIONS
    )

    nom = Faker('last_name')
    prenom = Faker('first_name')
    tel = Faker('phone_number')

    class Meta:
        model = Personne


class PersonneUserAccountFactory(UserAccountFactory):
    permissions__codenames = PERSONNE_PERMISSIONS
    personne = factory.RelatedFactory(PersonneFactory, 'account')
