from django.core.exceptions import ValidationError

import pytest

from ..models import AbstractUserType, UserAccount, UserTypeInfo
from .factories import OrganizationFactory, PersonFactory, UserAccountFactory
from .models import Organization, Person


@pytest.mark.django_db
class TestUserAccount:
    def test_email(self):
        account = UserAccountFactory.build(email='Adresse@Example.Org')
        account.clean()
        # cf. django.contrib.auth.base_user.BaseUserManager.normalize_email
        assert account.email == 'Adresse@example.org'

    def test_email_user(self, mailoutbox):
        account = UserAccountFactory.build()
        account.email_user('sujet', 'body')
        assert len(mailoutbox) == 1

    def test_is_staff(self):
        account = UserAccountFactory.build(is_superuser=False)
        assert account.is_staff is False
        account.is_superuser = True
        assert account.is_staff is True

    def test_names(self):
        person = PersonFactory()
        account = UserAccount.objects.get(pk=person.account_id)
        assert account.get_full_name() == person.get_full_name()
        assert account.get_short_name() == person.get_short_name()

    def test_details(self):
        account = UserAccountFactory()
        assert account.details is None

        person = PersonFactory()
        assert person.account.details == person

    def test_refresh_user_type_with_details(self):
        account = UserAccountFactory()
        with pytest.raises(ValueError):
            account.refresh_user_type(details=UserAccountFactory.build())
        with pytest.raises(ValueError):
            account.refresh_user_type(details=PersonFactory.build())

    def test_refresh_user_type(self):
        account = UserAccountFactory.build()
        account.refresh_user_type()
        assert account.details is None

        account.save()
        account.refresh_user_type()
        assert account.details is None

        organization = OrganizationFactory.build(account=account)
        account.refresh_user_type()
        assert account.details is None

        organization.save()
        account.refresh_user_type()
        assert account.details == organization


@pytest.mark.django_db
class TestUserType:
    def test_type_info(self):
        meta = Person._meta
        type_info = Person.type_info
        # vérifie les attributs générés
        assert isinstance(type_info, UserTypeInfo)
        assert type_info.name == meta.model_name
        assert type_info.verbose_name == meta.verbose_name
        assert type_info.verbose_name_plural == meta.verbose_name_plural

        person = PersonFactory()
        # vérifie l'accès aux infos depuis le compte
        assert person.account.type_info == type_info
        account = UserAccount.objects.get(pk=person.account_id)
        assert account.type_info == type_info

    def test_type_info_protected(self):
        with pytest.raises(AttributeError):

            class UserType(AbstractUserType):
                type_info = 'info'

    def test_validate_account(self):
        account = UserAccountFactory()
        organization = OrganizationFactory()
        org_account = organization.account

        # empêche l'utilisation d'un compte déjà associé
        with pytest.raises(ValueError):
            OrganizationFactory.build(account=org_account).save()
        with pytest.raises(ValidationError) as excinfo:
            OrganizationFactory.build(account=org_account).clean()
        assert 'account' in excinfo.value.error_dict.keys()

        # empêche le changement du compte d'un utilisateur
        with pytest.raises(ValueError):
            organization.account = account
            organization.save()
        with pytest.raises(ValidationError) as excinfo:
            organization.account = account
            organization.clean()
        assert 'account' in excinfo.value.error_dict.keys()

        # définir le même compte ne bloque pas
        organization.account = org_account
        organization.clean()

    def test_clean(self):
        Organization(name="Org").clean()
