import re

from django.contrib.messages import get_messages
from django.urls import reverse

import pytest

from boiteaasso.accounts.models import UserAccount
from boiteaasso.accounts.tests_accounts.factories import (
    AssociationUserAccountFactory,
    PersonneUserAccountFactory,
)
from boiteaasso.utils.tests import count_text_in_content, log_in_client

# ASSOCIATION
# -----------

# * tester : demande d'un nouveau jeton d'enregistrement


@pytest.mark.django_db(transaction=True)
class TestRegisterAssociation:
    def setup_method(self):
        self.url = reverse(
            'accounts:registration_register', args=['association']
        )

    def test_post_with_long_name(self, client, mailoutbox):
        data = {
            'account-email': 'blang@bling.org',
            'association-code_postal': '62000',
            'association-responsable_nom': 'Bling',
            'association-responsable_email': 'bling@blang.org',
            'association-activite_type': 'loisirs',
            'association-geolocalisation': '50,2.7',
            'association-responsable_prenom': 'Bling',
            'account-password1': 'xxxxxxx12;;',
            'account-password2': 'xxxxxxx12;;',
            'association-ville': 'Plop',
            'association-nom': 'Lorem ipsum dolor sit amet, consectetur',
            'association-responsable_fonction': 'présidente',
            'association-responsable_tel': '01 23 45 67 89',
            'association-sigle': '',
            'association-rue': '1 boulevard Lapointe ',
        }
        response = client.post(self.url, data)
        assert response.status_code == 302

        # la page de réussite est appelée
        assert response['Location'] == reverse('accounts:registration_complete')

        # un courriel est envoyé au demandeur
        assert len(mailoutbox) == 1
        assert "Activation de votre compte" in mailoutbox[0].subject

        # le nom n'est pas tronqué
        assert UserAccount.objects.last().short_name == data['association-nom']

        # un admin peut inspecter le compte
        admin = AssociationUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse('accounts_association_modeladmin_inspect', args=[1])
        response = client.get(url)

        assert response.status_code == 200
        assert count_text_in_content(response, data['association-nom'])


# PERSONNE
# --------


@pytest.mark.django_db(transaction=True)
class TestRegisterPersonne:
    def setup_method(self):
        self.url = reverse('accounts:registration_register', args=['personne'])

    def test_post_with_long_name(self, client, mailoutbox):
        data = {
            'account-email': 'blang@bling.org',
            'account-password1': 'xxxxxxx12;;',
            'account-password2': 'xxxxxxx12;;',
            'personne-prenom': 'Lorem ipsum dolor sit amet, consectetur',
        }
        response = client.post(self.url, data)
        assert response.status_code == 302

        # la page de réussite est appelée
        assert response['Location'] == reverse('accounts:registration_complete')

        # un courriel est envoyé au demandeur
        assert len(mailoutbox) == 1
        assert "Activation de votre compte" in mailoutbox[0].subject

        # le nom n'est pas tronqué
        assert UserAccount.objects.last().short_name == data['personne-prenom']

        # un admin peut inspecter le compte
        admin = PersonneUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse('accounts_personne_modeladmin_inspect', args=[1])
        response = client.get(url)

        assert response.status_code == 200
        assert count_text_in_content(response, data['personne-prenom'])


# PERTE DE MOT DE PASSE
# ---------------------


@pytest.mark.django_db(transaction=True)
class TestResetPassword:
    def setup_method(self):
        self.url = reverse('accounts:password_reset')

    def test_post_password_reset(self, client, mailoutbox):
        person = PersonneUserAccountFactory(is_superuser=False)
        data = {'email': person.email}
        response = client.post(self.url, data)
        assert response.status_code == 302

        # la page de réussite est appelée
        assert response['Location'] == reverse('accounts:login')

        # un courriel est envoyé au demandeur
        assert len(mailoutbox) == 1
        assert "Réinitialisation de votre mot de passe" in mailoutbox[0].subject

        # on choppe l'url
        match = re.search(
            r'[^\s]+/accounts/password/reset/confirm/[^\s]+',
            mailoutbox[0].body,
        )
        assert match
        url = match.group(0)

        # on valide le jeton
        response = client.get(url)
        assert response.status_code == 302
        url = response['Location']

        # on défini le pass
        data = {
            'new_password1': 'new_p4ssw0rdz',
            'new_password2': 'new_p4ssw0rdz',
        }
        response = client.post(url, data)
        assert response.status_code == 302

        # on arrive sur la page de login
        assert response['Location'] == reverse('accounts:login')

        # on tente la connexion
        url = response['Location']
        data = {
            'username': person.email,
            'password': 'new_p4ssw0rdz',
        }
        response = client.post(url, data)  # , follow=True)
        assert (
            "Vous êtes désormais connecté-e."
            in list(get_messages(response.wsgi_request))[-1].message
        )
