from django.contrib import messages
from django.contrib.auth import views as auth_views
from django.http import Http404, HttpResponseRedirect
from django.urls import reverse, reverse_lazy

from registration.backends.default import views as registration_views
from wagtail.contrib.modeladmin.views import EditView

from boiteaasso.utils.modeladmin import RelatedInspectView

from .forms import (
    AssociationRegistrationForm,
    AuthenticationForm,
    PasswordResetForm,
    PersonneRegistrationForm,
    ResendActivationForm,
    SetPasswordForm,
)

# CONNEXION ET GESTION DU COMPTE
# ------------------------------------------------------------------------------


class LoginView(auth_views.LoginView):
    form_class = AuthenticationForm
    template_name = 'accounts/login.html'

    def get_success_url(self):
        messages.success(self.request, "Vous êtes désormais connecté-e.")
        return super().get_success_url()


class LogoutView(auth_views.LogoutView):
    def get_next_page(self):
        messages.success(self.request, "Vous avez été déconnecté.")
        return super().get_next_page()


class PasswordResetView(auth_views.PasswordResetView):
    form_class = PasswordResetForm
    success_url = reverse_lazy('accounts:login')
    email_template_name = 'registration/password_reset_email.txt'
    html_email_template_name = 'registration/password_reset_email.html'

    def form_valid(self, form):
        self.to_email = form.cleaned_data['email']
        return super().form_valid(form)

    def get_success_url(self):
        messages.success(
            self.request,
            (
                "Un courriel a été envoyé à <i>{}</i>, si cette adresse mail "
                "est valide. Il contient les instructions à suivre pour "
                "réinitialiser votre mot de passe.".format(self.to_email)
            ),
        )
        return super().get_success_url()


class PasswordResetConfirmView(auth_views.PasswordResetConfirmView):
    form_class = SetPasswordForm
    success_url = reverse_lazy('accounts:login')

    def get_success_url(self):
        messages.success(self.request, "Votre mot de passe a été réinitialisé.")
        return super().get_success_url()


# ENREGISTREMENT ET ACTIVATION DU COMPTE
# ------------------------------------------------------------------------------

USER_TYPES_REGISTRATION_FORM = {
    'association': AssociationRegistrationForm,
    'personne': PersonneRegistrationForm,
}


class ResendActivationView(registration_views.ResendActivationView):
    form_class = ResendActivationForm
    success_url = reverse_lazy('accounts:login')

    def render_form_submitted_template(self, form):
        """Ajoute un message et redirige vers la page de connexion."""
        messages.success(
            self.request,
            (
                "Un courriel a été envoyé à <i>{}</i>, si cette adresse mail "
                "est valide. Il contient un nouveau lien permettant d'activer "
                "votre compte.".format(form.cleaned_data['email'])
            ),
        )
        return HttpResponseRedirect(self.get_success_url())


class ActivationView(registration_views.ActivationView):
    def activate(self, *args, **kwargs):
        user = super().activate(*args, **kwargs)
        if user and user.is_active:
            # tente d'ajouter automatiquement le nouvel utilisateur dans le
            # groupe de son type, s'il existe
            user.add_to_group_for_type()
        return user

    def get_success_url(self, user):
        messages.success(
            self.request,
            (
                "Votre compte est désormais activé ! Vous pouvez maintenant "
                "vous connecter sur la plateforme."
            ),
        )
        return ('accounts:login', (), {})


class RegistrationView(registration_views.RegistrationView):
    def dispatch(self, request, type_name):
        if type_name not in USER_TYPES_REGISTRATION_FORM:
            raise Http404(
                "Type d'utilisateur inconnu pour la création de compte."
            )
        self.type_name = type_name
        self.form_class = USER_TYPES_REGISTRATION_FORM[type_name]
        return super().dispatch(request)

    def get_template_names(self):
        return [
            'registration/{}/registration_form.html'.format(self.type_name),
            'registration/registration_form.html',
        ]


# ADMINISTRATION DES COMPTES
# ------------------------------------------------------------------------------


class MetaTitleMixin:
    def get_meta_title(self):
        """Retourne la propriété `meta_title` si elle existe."""
        return getattr(self, 'meta_title', super().get_meta_title())


class UserTypeEditView(MetaTitleMixin, EditView):
    meta_title = "Modification du compte"

    def get_success_url(self):
        if self.instance.account == self.request.user:
            return reverse('wagtailadmin_account')
        return super().get_success_url()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # ajoute la permission de l'utilisateur sur les comptes au contexte
        context['user_can_edit_account'] = self.request.user.has_perm(
            'accounts.useraccount.edit'
        )
        return context


class UserTypeInspectView(MetaTitleMixin, RelatedInspectView):
    page_title = "Détails"
    meta_title = "Détails sur le compte"


class AssociationEditView(UserTypeEditView):
    meta_title = "Modification de l'association"


class AssociationInspectView(UserTypeInspectView):
    meta_title = "Détails sur l'association"


class PersonneEditView(UserTypeEditView):
    meta_title = "Modification de la personne"


class PersonneInspectView(UserTypeInspectView):
    meta_title = "Détails sur la personne"
