from django.urls import reverse
from django.utils.html import format_html

from wagtail import hooks
from wagtail.contrib.modeladmin.helpers import ButtonHelper, PermissionHelper
from wagtail.contrib.modeladmin.options import (
    ModelAdmin,
    ModelAdminGroup,
    modeladmin_register,
)

from .models import Association, Personne
from .views import (
    AssociationEditView,
    AssociationInspectView,
    PersonneEditView,
    PersonneInspectView,
)


@hooks.register('register_account_menu_item')
def register_account_details_update(request):
    if not request.user.type_info:
        return {}
    return {
        'url': reverse(
            'accounts_{}_modeladmin_edit'.format(request.user.type_info.name),
            args=[request.user.details.pk],
        ),
        'label': "Détails du compte",
        'help_text': (
            "Changer les détails de votre compte {}.".format(
                request.user.type_info.verbose_name.title()
            )
        ),
    }


class UserTypePermissionHelper(PermissionHelper):
    """
    Permissions pour l'administration d'un type d'utilisateurs.
    """

    def user_can_create(self, user):
        # désactive la création depuis l'admin.
        return False

    def user_can_edit_obj(self, user, obj):
        return (
            hasattr(obj, 'account') and obj.account == user
        ) or super().user_can_edit_obj(user, obj)


class UserTypeButtonHelper(ButtonHelper):
    """
    Boutons pour l'administration d'un type d'utilisateurs.
    """

    def inspect_button(self, *args, **kwargs):
        button = super().inspect_button(*args, **kwargs)
        # modifie le titre du bouton avec les attributs de la vue
        inspect_view_class = self.view.model_admin.inspect_view_class
        if inspect_view_class:
            button['label'] = inspect_view_class.page_title
            button['title'] = inspect_view_class.meta_title
        return button


class AssociationsModelAdmin(ModelAdmin):
    """
    Administration des associations.
    """

    model = Association
    menu_label = "Associations"
    menu_icon = 'group'
    add_to_settings_menu = True
    list_display = ('nom_formatted', 'get_adresse_display')
    list_filter = ('activite_type', 'account__is_active')
    search_fields = ('nom',)
    inspect_view_enabled = True
    edit_view_class = AssociationEditView
    inspect_view_class = AssociationInspectView
    inspect_view_fields = (
        # informations sur l'association
        'nom',
        'sigle',
        'account__email',
        'adresse',
        'activite_type',
        # responsable de l'association
        'responsable_prenom',
        'responsable_nom',
        'responsable_fonction',
        'responsable_email',
        'responsable_tel',
        # détails sur le compte
        'account__is_active',
        'account__date_joined',
        'account__last_login',
        # statistiques sur les petites annonces
        'account__annonce_stats__annonces_nb',
        'account__annonce_stats__messages_nb',
    )
    button_helper_class = UserTypeButtonHelper
    permission_helper_class = UserTypePermissionHelper

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'nom_formatted':
            return ['title']
        return []

    # Mise en forme des champs ------------------------------------------------

    def nom_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette association">{}</a></h2>',
            self.url_helper.get_action_url('inspect', obj.pk),
            obj.nom,
        )

    nom_formatted.short_description = "nom"
    nom_formatted.admin_order_field = 'nom'


class PersonnesModelAdmin(ModelAdmin):
    """
    Administration des associations.
    """

    model = Personne
    menu_label = "Personnes"
    menu_icon = 'user'
    add_to_settings_menu = True
    list_display = ('nom_formatted', 'tel', 'get_adresse_display')
    list_filter = ('account__is_active',)
    search_fields = ('nom', 'prenom')
    inspect_view_enabled = True
    edit_view_class = PersonneEditView
    inspect_view_class = PersonneInspectView
    inspect_view_fields = (
        # informations sur la personne
        'nom',
        'prenom',
        'account__email',
        'tel',
        'adresse',
        # détails sur le compte
        'account__is_active',
        'account__date_joined',
        'account__last_login',
        # statistiques sur les petites annonces
        'account__annonce_stats__annonces_nb',
        'account__annonce_stats__messages_nb',
    )
    button_helper_class = UserTypeButtonHelper
    permission_helper_class = UserTypePermissionHelper

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'nom_formatted':
            return ['title']
        return []

    # Mise en forme des champs ------------------------------------------------

    def nom_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette personne">{}</a></h2>',
            self.url_helper.get_action_url('inspect', obj.pk),
            obj.get_full_name(),
        )

    nom_formatted.short_description = "nom"
    nom_formatted.admin_order_field = 'nom'


@modeladmin_register
class AccountsAdminGroup(ModelAdminGroup):
    menu_label = "Comptes"
    menu_icon = 'user'
    items = (AssociationsModelAdmin, PersonnesModelAdmin)
