from smtplib import SMTPException

from django.conf import settings
from django.contrib import messages
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string


class MapViewMixin:
    """
    Mixin pour afficher une carte Leaflet dans une vue, héritant déjà de
    `PageMixin` afin de prendre en compte les médias.
    """

    class Media:
        css = {
            'all': (
                'vendor/leaflet/leaflet.css',
                'vendor/leaflet.markercluster/MarkerCluster.css',
                'vendor/leaflet.markercluster/MarkerCluster.Default.css',
            )
        }
        js = (
            'vendor/leaflet/leaflet.js',
            'vendor/leaflet.markercluster/leaflet.markercluster.js',
            'js/map{}.js'.format('.min' if not settings.DEBUG else ''),
        )

    def get_map_options(self, context):
        return {}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['map_options'] = self.get_map_options(context)
        return context


class ConfirmationEmailMixin:
    """
    Mixin pour envoyer un email de confirmation après avoir modifié un objet.
    """

    email_template = None
    email_template_html = None
    email_template_subject = None

    def send_mail(self, context, from_email, to_email, **email_kwargs):
        """Envoi un courriel avec les paramétres donnés."""
        subject = render_to_string(self.email_template_subject, context)
        # l'objet ne doit pas contenir de nouvelles lignes
        subject = ''.join(subject.splitlines())
        message_txt = render_to_string(self.email_template, context)
        message_html = render_to_string(self.email_template_html, context)

        # construit et envoi le courriel
        email = EmailMultiAlternatives(
            subject, message_txt, from_email, [to_email], **email_kwargs
        )
        email.attach_alternative(message_html, 'text/html')
        try:
            email.send()
        except SMTPException:
            m = (
                "L'envoi du message a échoué. Vérifier l'exactitude de "
                "l'adresse destinataire et/ou contactez un "
                "administrateur de la plateforme."
            )
            messages.error(self.request, m)

    def send_confirmation_email(self, to, context={}, **kwargs):
        default_context = {
            'object': getattr(self, 'object', None)
            or getattr(self, 'instance', None),
            'to': to,
            'user': self.request.user,
            'help_url': getattr(self, 'help_url', None),
            'base_url': "{}://{}".format(
                self.request.scheme, self.request.get_host()
            ),
        }
        default_context.update(context)
        self.send_mail(
            default_context, settings.DEFAULT_FROM_EMAIL, to.email, **kwargs
        )
