from django.core.exceptions import ValidationError
from django.db import models
from django.utils.html import format_html

from location_field.models.plain import PlainLocationField
from wagtail.admin.panels import FieldPanel, FieldRowPanel
from wagtail.fields import RichTextField
from wagtail.models import Page
from wagtail.search import index


class AbstractAdresseMixin:
    @property
    def has_adresse(self):
        """Détermine si l'adresse est définie."""
        return all([self.rue, self.ville, self.code_postal])

    @property
    def is_geolocated(self):
        """Détermine si l'adresse est géolocalisée."""
        return True if self.geolocalisation else False

    @property
    def latitude(self):
        """Retourne la latitude si l'adresse est géolocalisée."""
        return self.geolocalisation.split(',')[0] if self.is_geolocated else ''

    @property
    def longitude(self):
        """Retourne la longitude si l'adresse est géolocalisée."""
        return self.geolocalisation.split(',')[1] if self.is_geolocated else ''

    @property
    def adresse(self):
        """Retourne l'adresse complète formatée sur une ligne."""
        if not self.has_adresse:
            return ''
        return "{}, {} {}".format(self.rue, self.code_postal, self.ville)

    def get_adresse_display(self):
        """Retourne l'adresse complète formatée en HTML sur deux lignes."""
        if not self.has_adresse:
            return ''
        return format_html(
            "{}<br>{} {}", self.rue, self.code_postal, self.ville
        )

    get_adresse_display.short_description = "adresse"

    def clear_adresse(self):
        """Vide l'adresse et la géolocalisation actuelles."""
        self.rue = ''
        self.ville = ''
        self.code_postal = ''
        self.geolocalisation = ''

    # Propriétés de Wagtail -----------------------------------------------

    adresse_panels = [
        FieldPanel('rue'),
        FieldRowPanel(
            [
                FieldPanel('code_postal', classname='col4'),
                FieldPanel('ville', classname='col8'),
            ]
        ),
        FieldPanel('geolocalisation'),
    ]


class AdresseMixin(models.Model, AbstractAdresseMixin):
    rue = models.CharField("rue", max_length=100)
    ville = models.CharField("ville", max_length=100)
    code_postal = models.CharField("code postal", max_length=5)
    geolocalisation = PlainLocationField(
        based_fields=['rue', 'ville'],
        blank=True,
        verbose_name="géolocalisation",
        help_text=(
            "Le marqueur est initié en fonction de l'adresse, déplacez le "
            "pour affiner la position. Vous pouvez zoomer ou dézoomer avec "
            "la molette de la souris."
        ),
    )

    class Meta:
        abstract = True


class OptionalAdresseMixin(models.Model, AbstractAdresseMixin):
    rue = models.CharField("rue", max_length=100, blank=True)
    ville = models.CharField("ville", max_length=100, blank=True)
    code_postal = models.CharField("code postal", max_length=5, blank=True)
    geolocalisation = PlainLocationField(
        based_fields=['rue', 'ville'],
        blank=True,
        verbose_name="géolocalisation",
        help_text=(
            "Le marqueur est initié en fonction de l'adresse, déplacez le "
            "pour affiner la position. Vous pouvez zoomer ou dézoomer avec "
            "la molette de la souris."
        ),
    )

    class Meta:
        abstract = True

    def clean(self):
        if not self.has_adresse:
            if any((self.rue, self.ville, self.code_postal)):
                raise ValidationError(
                    "L'adresse est renseignée mais n'est pas complète."
                )
            else:
                self.geolocalisation = ''


class OwnerLocalisationMixin(models.Model):
    class Meta:
        abstract = True

    use_owner_geolocalisation = models.BooleanField(
        "localisation du ou de la propriétaire",
        default=False,
        help_text=(
            "Détermine s'il faut utiliser la localisation du ou de la"
            "propriétaire à la place, si elle est définie."
        ),
    )

    @property
    def has_adresse(self):
        if self.use_owner_geolocalisation:
            return getattr(self.owner.details, 'has_adresse', False)
        return super().has_adresse

    @property
    def is_geolocated(self):
        if self.use_owner_geolocalisation:
            return getattr(self.owner.details, 'is_geolocated', False)
        return super().is_geolocated

    @property
    def latitude(self):
        if self.use_owner_geolocalisation:
            return getattr(self.owner.details, 'latitude', '')
        return super().latitude

    @property
    def longitude(self):
        if self.use_owner_geolocalisation:
            return getattr(self.owner.details, 'longitude', '')
        return super().longitude

    @property
    def adresse(self):
        if self.has_adresse and self.use_owner_geolocalisation:
            return self.owner.details.adresse
        return super().adresse

    def get_adresse_display(self):
        if self.has_adresse and self.use_owner_geolocalisation:
            return self.owner.details.get_adresse_display()
        return super().get_adresse_display()

    def clean(self):
        if self.use_owner_geolocalisation:
            if (
                not hasattr(self, 'owner')
                or self.owner.details is None
                or not hasattr(self.owner.details, 'geolocalisation')
            ):
                raise ValidationError(
                    {
                        'use_owner_geolocalisation': (
                            "L'adresse n'est pas géolocalisable, "
                            "veuillez saisir l'adresse ci-dessous à la place."
                        )
                    }
                )

            # pour éviter les confusions, on vide l'adresse
            self.clear_adresse()

        super().clean()


# MODÈLES DE PAGES
# ------------------------------------------------------------------------------


class HomePage(Page):
    """
    La page d'accueil de la plateforme, qui pourra évoluer par la suite.
    """

    body = RichTextField("contenu")

    class Meta:
        verbose_name = "page d'accueil"
        verbose_name_plural = "pages d'accueil"

    # Propriétés de Wagtail ---------------------------------------------------

    search_fields = Page.search_fields + [index.SearchField('body')]

    content_panels = Page.content_panels + [
        FieldPanel('body', classname='full')
    ]

    parent_page_types = ['wagtailcore.Page']


class SimplePage(Page):
    """
    Une page simple avec uniquement du contenu enrichi.
    """

    body = RichTextField("contenu")

    class Meta:
        verbose_name = "page simple"
        verbose_name_plural = "pages simples"

    # Propriétés de Wagtail ---------------------------------------------------

    search_fields = Page.search_fields + [index.SearchField('body')]

    content_panels = Page.content_panels + [
        FieldPanel('body', classname='full')
    ]

    show_in_menus_default = True
