from datetime import date, datetime

from django import template
from django.template import defaultfilters
from django.utils.safestring import mark_safe

register = template.Library()


def format_date(value, date_format=None):
    """
    Ce filtre est basé sur ``naturalday`` de ``django.contrib.humanize``.
    En plus d'ajouter l'heure, il vérifie en plus s'il est utile d'ajouter
    l'année ou pas. Il est juste possible de personnaliser le format de la
    date, dans le cas où elle n'est pas "aujourd'hui", "demain" ou "hier".
    L'année sera automatiquement enlevée si elle n'est pas utile.
    """
    date_format = date_format or 'j F Y'
    try:
        tzinfo = getattr(value, 'tzinfo', None)
        value = date(value.year, value.month, value.day)
    except AttributeError:
        # Passed value wasn't a date object
        return value
    except ValueError:
        # Date arguments out of range
        return value
    today = datetime.now(tzinfo).date()
    delta = value - today
    if delta.days == 0:
        date_string = "aujourd'hui"
    elif delta.days == 1:
        date_string = "demain"
    elif delta.days == -1:
        date_string = "hier"
    elif delta.days >= 365:
        date_string = defaultfilters.date(value, date_format)
    else:
        date_string = defaultfilters.date(
            value, date_format.replace('Y', '').replace('y', '').strip()
        )
    return date_string


@register.filter(expects_localtime=True)
def postdate(value, date_format=None):
    """
    Formate une date et une heure pour être plus parlant pour un humain,
    en remplaçant par exemple la date par "aujourd'hui" si c'est le cas.
    """
    date_string = format_date(value, date_format)
    iso_datetime = defaultfilters.date(value, 'c')
    return mark_safe(
        '<time datetime="{0}">{1}</time>'.format(
            iso_datetime, date_string.lower()
        )
    )


@register.filter(expects_localtime=True)
def postdatetime(value, date_format=None):
    """
    Formate une date et une heure pour être plus parlant pour un humain,
    en remplaçant par exemple la date par "aujourd'hui" si c'est le cas.
    Le texte est englobé dans un élément ``<time>``.
    """
    date_string = format_date(value, date_format)
    time_string = defaultfilters.time(value)
    iso_datetime = defaultfilters.date(value, 'c')
    return mark_safe(
        '<time datetime="{0}">{1} à {2}</time>'.format(
            iso_datetime, date_string.lower(), time_string
        )
    )
