from django import forms

from tapeforms.contrib.foundation import FoundationTapeformMixin

from .models import AnnonceCategory


class AnnoncesFiltersForm(FoundationTapeformMixin, forms.Form):
    """
    Formulaire de filtres sur les petites annonces.
    """

    category = forms.ModelChoiceField(
        queryset=AnnonceCategory.objects,
        empty_label="Toutes les catégories",
        required=False,
    )
    query = forms.CharField(
        required=False,
        widget=forms.TextInput(attrs={'placeholder': "Mots clés…"}),
    )

    field_template = 'tapeforms/fields/simple.html'

    @property
    def has_filters(self):
        """
        Détermine si le formulaire a des critères de définis.
        """
        return self.is_bound and (
            self.data.get('category') or self.data.get('query')
        )

    def filter_queryset(self, queryset):
        """Filtre la sélection avec les critères du formulaire.

        Retourne une nouvelle sélection de petites annonces à partir de
        `queryset` en la filtrant avec les valeurs du formulaire, après
        l'avoir validé. Si le critère `query` est définis, la sélection
        retournée sera une instance ``SearchResults`` de Wagtail au lieu
        d'une ``QuerySet``.
        """
        category = self.cleaned_data.get('category')
        if category:
            queryset = queryset.filter(category=category)
        query = self.cleaned_data.get('query')
        if query:
            queryset = queryset.search(query, order_by_relevance=False)
        return queryset


class AnnonceContactForm(FoundationTapeformMixin, forms.Form):
    """
    Formulaire de contact sur une petite annonce.
    """

    message = forms.CharField(
        min_length=5,
        error_messages={
            'required': "Le contenu du message est requis.",
            'min_length': "Le contenu du message est trop court.",
        },
        label="Votre message",
        widget=forms.Textarea,
    )
