import logging
import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.text import slugify

from djchoices import ChoiceItem, DjangoChoices
from dynamic_filenames import FilePattern
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from stdimage.models import StdImageField
from wagtail.admin.panels import (
    FieldPanel,
    HelpPanel,
    InlinePanel,
    MultiFieldPanel,
    ObjectList,
)
from wagtail.contrib.settings.models import BaseSiteSetting, register_setting
from wagtail.fields import RichTextField
from wagtail.models import Orderable
from wagtail.search import index
from wagtail.search.backends import get_search_backend

from boiteaasso.accounts.models import UserAccount
from boiteaasso.base.models import OptionalAdresseMixin, OwnerLocalisationMixin
from boiteaasso.base.widgets import AdminImageWidget
from boiteaasso.utils.functional import cached_class_property

logger = logging.getLogger(__name__)

"""Le nombre maximum de photos sur une petite annonce."""
ANNONCE_PHOTOS_MAX = 3

"""La taille de la vignette d'une photo d'une petite annonce."""
ANNONCE_THUMB_SIZE = 150


class AnnonceCategory(models.Model):
    """
    Une catégorie utilisée pour qualifier une petite annonce.
    """

    name = models.CharField("nom", max_length=255)
    slug = models.SlugField(unique=True, max_length=80)

    class Meta:
        ordering = ['name']
        verbose_name = "catégorie"
        verbose_name_plural = "catégories des petites annonces"

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # génère le slug depuis le nom
        self.slug = slugify(self.name)
        return super().save(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [FieldPanel('name', classname='title')]


class AnnonceType(DjangoChoices):
    """
    Les types de petites annonces disponibles.
    """

    offre = ChoiceItem('offre', "Offre", label_plural="Offres")
    demande = ChoiceItem('demande', "Demande", label_plural="Demandes")


class AnnonceQuerySet(models.QuerySet):
    def active(self):
        """Filtre avec les petites annonces visibles publiquement."""
        return self.filter(is_active=True)

    def search(self, query, **kwargs):
        """Recherche une petite annonce avec des mots clés."""
        backend = get_search_backend()
        kwargs.setdefault('fields', ['title', 'description'])
        return backend.search(query, self, **kwargs)


class Annonce(
    ClusterableModel,
    OwnerLocalisationMixin,
    OptionalAdresseMixin,
    index.Indexed,
):
    """
    Une petite annonce, que ce soit une offre ou une demande.
    """

    title = models.CharField("titre", max_length=255)
    type = models.CharField("type", max_length=10, choices=AnnonceType.choices)
    category = models.ForeignKey(
        AnnonceCategory,
        on_delete=models.SET_NULL,
        null=True,
        verbose_name="catégorie",
    )
    is_active = models.BooleanField(
        "active",
        default=True,
        help_text="Détermine si la petite annonce est visible publiquement.",
    )
    description = RichTextField(
        features=['h2', 'h3', 'bold', 'italic', 'link', 'ul']
    )

    # Localisation
    use_owner_geolocalisation = models.BooleanField(
        "localisation de l'annonceur-se",
        default=False,
        help_text=(
            "Détermine s'il faut utiliser la localisation de l'annonceur-se "
            "à la place, si elle est définie."
        ),
    )

    # Méta-données
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='annonces',
        editable=True,
        verbose_name="annonceur-se",
    )
    created_at = models.DateTimeField(
        "date de création", default=timezone.now, blank=True, editable=False
    )
    last_modified_at = models.DateTimeField(
        "date de la dernière modification", blank=True, editable=False
    )

    objects = AnnonceQuerySet.as_manager()

    class Meta:
        verbose_name = "petite annonce"
        verbose_name_plural = "petites annonces"
        permissions = (
            ('contact_annonce', "Peut contacter les annonceurs-ses"),
        )

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if self.last_modified_at is None or self.pk is not None:
            self.last_modified_at = timezone.now()
        return super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse('echanges:annonce-detail', args=[self.pk])

    def permissions_for_user(self, user):
        return AnnoncePermissionTester(self, user)

    # Gestion des médias associés ---------------------------------------------

    @property
    def has_photos(self):
        """
        Détermine si des photos sont définies pour cette petite annone.
        """
        return self.photos.count()

    def get_highlighted_photo(self):
        """
        Récupère, si elle existe, la photo mise en avant pour cette petite
        annonce. C'est actuellement la première photo qui est retournée.
        """
        return self.photos.first()

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        MultiFieldPanel(
            [
                FieldPanel('title'),
                FieldPanel('type'),
                FieldPanel('category'),
                FieldPanel('is_active'),
            ],
            heading="Petite annonce",
        ),
        MultiFieldPanel(
            [
                FieldPanel('description'),
                InlinePanel(
                    'photos',
                    max_num=ANNONCE_PHOTOS_MAX,
                    label="Photos",
                    help_text=(
                        "Vous pouvez ajouter {0} photos, la première sera "
                        "mise en avant et utilisée pour présenter la petite "
                        "annonce.".format(
                            "plusieurs"
                            if not ANNONCE_PHOTOS_MAX
                            else "jusqu'à {}".format(ANNONCE_PHOTOS_MAX)
                        )
                    ),
                ),
            ],
            heading="Description",
        ),
        MultiFieldPanel(
            [FieldPanel('use_owner_geolocalisation')]
            + OptionalAdresseMixin.adresse_panels,
            heading="Localisation",
        ),
    ]

    search_fields = [
        index.SearchField('title', partial_match=True, boost=2),
        index.FilterField('type'),
        index.FilterField('category'),
        index.SearchField('description'),
        index.FilterField('is_active'),
        index.FilterField('owner'),
        index.FilterField('created_at'),
        index.FilterField('last_modified_at'),
    ]


class AnnoncePermissionTester:
    """
    Testeur des permissions d'un utilisateur sur une petite annonce.
    """

    perm_re = re.compile(r'echanges\.(\w+)_annonce')

    def __init__(self, annonce, user):
        self.annonce = annonce
        self.user = user

    @cached_property
    def permissions(self):
        """Retourne les permissions de l'utilisateur liées aux annonces."""
        permissions = set()
        for perm in self.user.get_all_permissions():
            match = self.perm_re.match(perm)
            if match:
                permissions.add(match.group(1))
        return permissions

    def is_owner(self):
        return self.annonce.owner_id == self.user.pk

    def can_edit(self):
        return (
            self.user.is_superuser
            or 'change' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_delete(self):
        return (
            self.user.is_superuser
            or 'delete' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_contact(self):
        return not self.is_owner() and 'contact' in self.permissions


class AnnoncePhoto(Orderable):
    """
    Une photo au sein d'une petite annonce.
    """

    annonce = ParentalKey(
        Annonce, on_delete=models.CASCADE, related_name='photos'
    )
    photo_upload_to = FilePattern(
        filename_pattern='annonces/photos/{uuid:x}{ext}'
    )
    photo = StdImageField(
        upload_to=photo_upload_to,
        variations={
            'thumbnail': {
                'width': ANNONCE_THUMB_SIZE,
                'height': ANNONCE_THUMB_SIZE,
            }
        },
    )
    title = models.CharField("titre", max_length=255, blank=True)

    class Meta:
        ordering = ('annonce', 'sort_order')
        verbose_name = "photo d'une annonce"
        verbose_name_plural = "photos d'une annonce"

    def __str__(self):
        return self.title or 'Untitled'

    def clean(self, *args, **kwargs):
        if not self.title:
            # définis le titre depuis le nom du fichier uniquement
            # pour une nouvelle photo (après, le nom est modifié)
            if self.photo.name and not self.photo._committed:
                self.title = self.photo.name.rsplit('.', 1)[0]
            else:
                raise ValidationError(
                    {'title': "Veuillez renseigner un titre pour cette photo."}
                )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        FieldPanel('photo', widget=AdminImageWidget),
        FieldPanel('title'),
    ]


class UserAnnonceStats(models.Model):
    """
    Statistiques sur les petites annonces pour un utilisateur-rice.
    """

    account = models.OneToOneField(
        UserAccount,
        on_delete=models.CASCADE,
        related_name='annonce_stats',
        primary_key=True,
    )
    annonces_nb = models.PositiveSmallIntegerField(
        "nombre d'annonces",
        default=0,
        help_text="Nombre total de petites annonces publiées.",
    )
    messages_nb = models.PositiveSmallIntegerField(
        "nombre de messages",
        default=0,
        help_text=(
            "Nombre total de messages envoyés sur les petites annonces."
        ),
    )

    @classmethod
    def increment(cls, counter, account):
        """Incrémente un compteur pour un utilisateur-rice."""
        obj, created = cls.objects.get_or_create(
            account=account, defaults={counter: 1}
        )
        if not created:
            setattr(obj, counter, getattr(obj, counter) + 1)
            obj.save()
        return obj


@register_setting(icon='bullhorn')
class EchangesSettings(BaseSiteSetting):
    """
    Paramètres du service d'échanges.
    """

    introduction = models.TextField(
        help_text=(
            "Un petit texte d'introduction pour décrire le service, affiché "
            "en haut de la page."
        )
    )
    last_annonces_nb = models.PositiveSmallIntegerField(
        "dernières petites annonces",
        default=3,
        help_text=(
            "Le nombre de dernières petites annonces à afficher, ou 0 pour "
            "les masquer."
        ),
    )
    help_is_active = models.BooleanField(
        "aide active",
        default=False,
        help_text="Définis si la page d'aide est visible.",
    )
    help_content = RichTextField("contenu de l'aide", blank=True)

    class Meta:
        verbose_name = "paramètres du service d'échanges"

    def __str__(self):
        return self.site.site_name or "Site par défaut"

    def clean(self, *args, **kwargs):
        if self.help_is_active and not self.help_content:
            raise ValidationError(
                {
                    'help_content': (
                        "Veuillez définir le contenu de la page ou la "
                        "désactiver."
                    )
                }
            )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    @cached_class_property
    def edit_handler(cls):
        category_url = 'echanges_annoncecategory_modeladmin_index'
        return ObjectList(
            [
                MultiFieldPanel(
                    [
                        HelpPanel(
                            content=format_html(
                                '<a href="{}" class="button">{}</a>',
                                reverse(category_url),
                                "Modifier les catégories",
                            ),
                            template=(
                                'wagtailadmin/edit_handlers/raw_panel.html'
                            ),
                        )
                    ],
                    heading="Petites annonces",
                ),
                MultiFieldPanel(
                    [
                        FieldPanel('introduction'),
                        FieldPanel('last_annonces_nb'),
                    ],
                    heading="Page d'accueil",
                ),
                MultiFieldPanel(
                    [FieldPanel('help_is_active'), FieldPanel('help_content')],
                    heading="Page d'aide",
                ),
            ]
        )
