from django.conf import settings
from django.contrib.messages import ERROR as ERROR_LEVEL
from django.contrib.messages import SUCCESS as SUCCESS_LEVEL
from django.contrib.messages import get_messages
from django.urls import reverse, reverse_lazy

import pytest
from bs4 import BeautifulSoup as bs

from boiteaasso.accounts.tests_accounts.factories import (
    AssociationUserAccountFactory,
    PersonneUserAccountFactory,
)
from boiteaasso.utils.tests import (
    count_text_in_content,
    log_in_client,
    skipif_sqlite,
)

from ..models import AnnonceType, UserAnnonceStats
from .factories import (
    AnnonceCategoryFactory,
    AnnonceFactory,
    DemandeFactory,
    OffreFactory,
)


@pytest.mark.django_db
class TestAnnonceList:
    url = reverse_lazy('echanges:annonces')

    def test_get(self, client):
        annonces = AnnonceFactory.create_batch(4)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['annonces']) == len(annonces)

    def test_filter_type(self, client):
        offres = OffreFactory.create_batch(2)

        response = client.get(
            reverse(
                'echanges:annonces-by-type',
                kwargs={'annonces_type': AnnonceType.offre},
            )
        )
        assert len(response.context['annonces']) == len(offres)

        response = client.get(
            reverse(
                'echanges:annonces-by-type',
                kwargs={'annonces_type': AnnonceType.demande},
            )
        )
        assert len(response.context['annonces']) == 0

    def test_filter_category(self, client):
        category = AnnonceCategoryFactory()
        annonces = [
            OffreFactory(category=category),
            DemandeFactory(category=category),
        ] + DemandeFactory.create_batch(2)

        # la catégorie contient 2 annonces
        response = client.get(self.url, {'category': category.pk})
        assert response.status_code == 200
        assert len(response.context['annonces']) == 2

        # la catégorie est inexistante, on affiche tout
        response = client.get(self.url, {'category': 'nope'})
        assert response.status_code == 200
        assert len(response.context['annonces']) == len(annonces)

    @skipif_sqlite
    def test_filter_query(self, client):
        AnnonceFactory(title="Test azertyuiop")
        AnnonceFactory.create_batch(2)

        # des mots clés entiers
        response = client.get(self.url, {'query': "Test azertyuiop"})
        assert response.status_code == 200
        assert len(response.context['annonces']) == 1

        # un mot clé partiel
        response = client.get(self.url, {'query': "zertyui"})
        assert response.status_code == 200
        assert len(response.context['annonces']) == 1

        # un mot clé inexistant
        response = client.get(self.url, {'query': "nope"})
        assert response.status_code == 200
        assert len(response.context['annonces']) == 0


@pytest.mark.django_db
class TestAnnonceDetail:
    def setup_method(self):
        self.annonce = AnnonceFactory()
        self.url = self.annonce.get_absolute_url()

    def test_get_anonymous(self, client):
        response = client.get(self.url)
        assert response.status_code == 200
        assert not count_text_in_content(response, "Ajoutée par :")
        assert not count_text_in_content(response, "Contacter…")

        actions = response.context['page']['actions']
        assert len(actions) == 1
        assert actions[0]['url'] == reverse('echanges:annonces')
        assert actions[0]['title'] == "Voir les petites annonces"

    def test_get_owner(self, client):
        assert log_in_client(client, self.annonce.owner)

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "Ajoutée par :")
        assert not count_text_in_content(response, "Contacter…")

        actions = response.context['page']['actions']
        # l'utilisateur peut modifier sa petite annonce
        assert len(actions) == 2
        assert actions[1]['url'] == reverse(
            'echanges_annonce_modeladmin_edit', args=[self.annonce.pk]
        )
        assert actions[1]['title'] == "Modifier cette petite annonce"

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_get_other(self, client, logged_user):
        assert log_in_client(client, logged_user())

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "Ajoutée par :")
        assert count_text_in_content(response, "Contacter…")
        assert len(response.context['page']['actions']) == 1


@pytest.mark.django_db
class TestAnnonceDetailLocation:
    def test_no_adresse(self, client):
        annonce = AnnonceFactory()
        assert not annonce.has_adresse

        response = client.get(annonce.get_absolute_url())
        assert not count_text_in_content(response, "Localisation")

    def test_owner_adresse_not_geolocated(self, client):
        account = AssociationUserAccountFactory(
            association__rue="3 rue des coquelicots",
            association__ville="Libreville",
            association__code_postal="01000",
        )
        annonce = AnnonceFactory(owner=account, use_owner_geolocalisation=True)
        assert annonce.has_adresse
        assert not annonce.is_geolocated

        response = client.get(annonce.get_absolute_url())
        assert count_text_in_content(response, "Localisation")
        assert count_text_in_content(response, "Libreville")
        assert not count_text_in_content(response, "boiteaassoMapOptions")
        assert not response.context['map_options']

    def test_annonce_adresse_geolocated(self, client):
        annonce = AnnonceFactory(
            use_owner_geolocalisation=False,
            rue="3 rue des coquelicots",
            ville="Libreville",
            code_postal="01000",
            geolocalisation="1,2",
        )
        assert annonce.has_adresse
        assert annonce.is_geolocated

        response = client.get(annonce.get_absolute_url())
        assert count_text_in_content(response, "Localisation")
        assert count_text_in_content(response, "Libreville")
        assert count_text_in_content(response, "boiteaassoMapOptions")
        assert response.context['map_options']['markers'] == [
            {
                'lat': '1',
                'lng': '2',
                'popup': "3 rue des coquelicots<br>01000 Libreville",
            }
        ]


@pytest.mark.django_db
class TestAnnonceContact:
    def setup_method(self):
        self.annonce = AnnonceFactory()
        self.url = reverse('echanges:annonce-contact', args=[self.annonce.pk])

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_get(self, client, logged_user):
        assert log_in_client(client, logged_user())

        response = client.get(self.url)
        assert response.status_code == 200

    def test_post_anonymous(self, client):
        response = client.post(self.url)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse(settings.LOGIN_URL), self.url
        )

    def test_post_owner(self, client, mailoutbox):
        assert log_in_client(client, self.annonce.owner)

        response = client.post(
            self.url, {'message': "Ceci est mon message de test !"}
        )
        assert response.status_code == 302
        assert response['Location'] == self.annonce.get_absolute_url()

        # l'annonceur ne peut pas contacter son annonce
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == ERROR_LEVEL

        assert len(mailoutbox) == 0

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other_invalid(self, client, mailoutbox, logged_user):
        assert log_in_client(client, logged_user())

        response = client.post(self.url)
        assert response.status_code == 200

        # le message est requis
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == ERROR_LEVEL

        assert len(mailoutbox) == 0

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other(self, client, mailoutbox, logged_user):
        account = logged_user()
        assert log_in_client(client, account)

        response = client.post(
            self.url, {'message': "Ceci est mon message de test !"}
        )
        assert response.status_code == 302
        assert response['Location'] == self.annonce.get_absolute_url()

        # un message de réussite est affiché
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

        # un courriel est envoyé à l'annonceur
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [self.annonce.owner.email]

        # le compteur de l'association est incrémenté
        annonce_stats = UserAnnonceStats.objects.get(account=account)
        assert annonce_stats.messages_nb == 1


@pytest.mark.django_db
class TestHelpView:
    def test_get(self, client):
        url = reverse_lazy('echanges:help')
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_get_edit_form(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse_lazy(
            'wagtailsettings:edit', args=['echanges', 'echangessettings']
        )
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_follow_category_button(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse_lazy(
            'wagtailsettings:edit', args=['echanges', 'echangessettings']
        )
        response = client.get(url, follow=True)
        assert response.status_code == 200
        buttons = bs(response.content, 'html.parser').find_all(
            'a', class_='button'
        )
        assert len(buttons) > 0
        url = buttons[0].attrs['href']
        response = client.get(url)
        assert response.status_code == 200
