from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import AccessMixin
from django.http import Http404
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.views.generic import DetailView, ListView, TemplateView
from django.views.generic.edit import BaseFormView, FormMixin

from wagtail.contrib.modeladmin import views as modeladmin_views
from wagtail.models import Site

from boiteaasso.base.mixins import ConfirmationEmailMixin, MapViewMixin
from boiteaasso.utils.views import CollectionPageMixin

from .forms import AnnonceContactForm, AnnoncesFiltersForm
from .models import Annonce, AnnonceType, EchangesSettings, UserAnnonceStats


class SettingsMixin:
    @cached_property
    def settings(self):
        """Retourne les paramètres du service d'échanges."""
        return EchangesSettings.for_site(Site.find_for_request(self.request))

    @property
    def help_url(self):
        return (
            None
            if not self.settings.help_is_active
            else self.request.build_absolute_uri(reverse('echanges:help'))
        )


class EchangesPageMixin(CollectionPageMixin, SettingsMixin):
    collection_list_title = "Petites annonces"
    collection_list_url = reverse_lazy('echanges:annonces')
    show_breadcrumb = False

    def collection_include_list_crumb(self):
        return issubclass(self.__class__, DetailView)

    def get_breadcrumb(self):
        return [
            {'title': "Services"},
            {'title': "Échanges"},
        ] + super().get_breadcrumb()


class HelpView(EchangesPageMixin, TemplateView):
    template_name = 'echanges/help.html'
    page_title = "Aide pour les échanges"
    breadcrumb_title = "Aide"

    def get(self, request, *args, **kwargs):
        if not self.settings.help_is_active:
            return redirect(self.collection_list_url)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['content'] = self.settings.help_content
        return context

    def get_page_actions(self):
        return [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les petites annonces",
                'label': self.collection_list_title,
            }
        ]


class AnnonceListView(EchangesPageMixin, MapViewMixin, ListView):
    """
    Liste les petites annonces, éventuellement filtrées.
    """

    context_object_name = 'annonces'
    template_name = 'echanges/annonce_list.html'
    page_seo_title = "Liste des petites annonces"
    page_title = "Petites annonces"

    def get_queryset(self):
        """Filtre la sélection de petites annonces à afficher."""
        queryset = Annonce.objects.active().order_by('-last_modified_at')
        # filtre avec le type définis dans l'URL
        self.annonces_type = self.kwargs.get('annonces_type', None)
        if self.annonces_type is not None:
            try:
                self.annonces_type = AnnonceType.get_choice(self.annonces_type)
            except KeyError:
                raise Http404(
                    "Le type de petites annonces demandé est inconnu."
                )
            queryset = queryset.filter(type=self.annonces_type.value)
        # filtre avec les autres critères du formulaire
        self.filters_form = AnnoncesFiltersForm(self.request.GET)
        if self.filters_form.has_filters:
            self.filters_form.full_clean()
            queryset = self.filters_form.filter_queryset(queryset)
        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['annonces_type'] = self.annonces_type
        context['filters_form'] = self.filters_form
        context['introduction'] = self.settings.introduction
        return context

    def get_page_actions(self):
        actions = []
        if self.request.user.has_perm('echanges.add_annonce'):
            actions.append(
                {
                    'url': '{}?{}'.format(
                        reverse('echanges_annonce_modeladmin_index'),
                        AnnonceIndexView.USER_ANNONCES_VAR,
                    ),
                    'class': 'icon bullhorn',
                    'title': "Gérer mes annonces",
                    'label': "Mes annonces",
                }
            )
        if self.settings.help_is_active:
            actions.append(
                {
                    'url': reverse('echanges:help'),
                    'title': HelpView.page_title,
                    'label': mark_safe(
                        '<i class="fa fa-question-circle" aria-hidden="true">'
                        '</i>'
                    ),
                    'attrs': {'aria-label': "Aide sur les échanges"},
                }
            )
        return actions

    def get_map_options(self, context):
        return {
            'markers': self.get_annonces_markers(context['object_list']),
            'mapOptions': {'center': settings.DEFAULT_MAP_CENTER, 'zoom': 8},
        }

    def get_annonces_markers(self, object_list):
        return [
            {
                'lat': annonce.latitude,
                'lng': annonce.longitude,
                'popup': render_to_string(
                    'echanges/annonce_popup.html', {'annonce': annonce}
                ),
            }
            for annonce in object_list
            if annonce.is_geolocated
        ]


class ContactFormMixin(FormMixin):
    form_class = AnnonceContactForm


class AnnonceDetailView(
    EchangesPageMixin, MapViewMixin, ContactFormMixin, DetailView
):
    """
    Présente une petite annonce.
    """

    context_object_name = 'annonce'
    queryset = Annonce.objects.active()
    template_name = 'echanges/annonce.html'

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)
        # récupère les permissions de l'utilisateur sur l'annonce
        self.user_perms = obj.permissions_for_user(self.request.user)
        return obj

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['can_contact'] = self.user_perms.can_contact()
        context['form'] = self.get_form()
        return context

    def get_page_title(self):
        return self.object.title

    def get_page_seo_title(self):
        return '{} : {}'.format(
            self.object.get_type_display(), self.object.title
        )

    def get_page_actions(self):
        actions = [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les petites annonces",
                'label': self.collection_list_title,
            }
        ]
        if self.user_perms.can_edit():
            actions.append(
                {
                    'url': reverse(
                        'echanges_annonce_modeladmin_edit',
                        args=[self.object.pk],
                    ),
                    'class': 'icon fa-pencil',
                    'title': "Modifier cette petite annonce",
                    'label': "Modifier…",
                }
            )
        return actions

    def get_map_options(self, context):
        if not self.object.is_geolocated:
            return {}
        return {
            'markers': [
                {
                    'lat': self.object.latitude,
                    'lng': self.object.longitude,
                    'popup': self.object.get_adresse_display(),
                }
            ],
            'mapOptions': {
                'center': [self.object.latitude, self.object.longitude],
                'zoom': 8,
            },
        }


# VUES RESTREINTES
# ------------------------------------------------------------------------------


class AnnonceContactView(
    AccessMixin,
    ConfirmationEmailMixin,
    AnnonceDetailView,
    BaseFormView,
):
    """
    Contact l'annonceur-se d'une petite annonce.
    """

    email_template = 'echanges/emails/annonce_contact.txt'
    email_template_html = 'echanges/emails/annonce_contact.html'
    email_template_subject = 'echanges/emails/annonce_contact_subject.txt'

    def has_permission(self):
        return self.object.permissions_for_user(self.request.user).can_contact()

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_permission():
            if not request.user.is_authenticated:
                return self.handle_no_permission()
            messages.error(
                self.request,
                "Vous n'avez pas la permission de contacter cette annonce.",
            )
            return redirect(self.object)
        return super().dispatch(request, *args, **kwargs)

    # Traitement du formulaire ------------------------------------------------

    def form_valid(self, form):
        context = {
            'from': self.request.user,
            'message': form.cleaned_data['message'],
        }
        self.send_confirmation_email(
            self.object.owner, context, reply_to=[self.request.user.email]
        )

        # incrémente le compteur de la personne
        UserAnnonceStats.increment('messages_nb', self.request.user)

        messages.success(
            self.request,
            (
                "Votre message pour {} a été envoyé.".format(
                    self.object.owner.get_full_name() or "l'annonceur-se"
                )
            ),
        )

        return redirect(self.object)

    def form_invalid(self, form):
        errors = (
            set().union(*form.errors.values()).union(form.non_field_errors())
        )
        if not errors:
            message = "Impossible d'envoyer votre message."
        elif len(errors) == 1:
            message = "Impossible d'envoyer votre message : {}".format(
                errors.pop().lower()
            )
        else:
            message = "Impossible d'envoyer votre message :{}".format(
                '<br>&nbsp;- '.join(['', *errors])
            )
        messages.error(self.request, message)

        return super().form_invalid(form)


class AnnonceCreateView(modeladmin_views.CreateView):
    def get_instance(self):
        # définis l'annonceur-se avec la personne connectée
        instance = super().get_instance()
        instance.owner = self.request.user
        return instance


class AnnonceIndexView(modeladmin_views.IndexView):
    USER_ANNONCES_VAR = 'self'
    # ajoute un argument pour filtrer avec les annonces de la personne
    IGNORED_PARAMS = modeladmin_views.IndexView.IGNORED_PARAMS + (
        USER_ANNONCES_VAR,
    )
