from django.utils.html import format_html

from wagtail import hooks
from wagtail.admin.site_summary import SummaryItem
from wagtail.contrib.modeladmin.helpers import ButtonHelper, PermissionHelper
from wagtail.contrib.modeladmin.options import ModelAdmin

from boiteaasso.base.templatetags.horodated import postdatetime

from .models import Annonce
from .views import AnnonceCreateView, AnnonceIndexView


class AnnoncesSummaryItem(SummaryItem):
    """
    Affiche le nombre de petites annonces de la personne ou au total.
    """

    order = 900
    template_name = 'echanges/dashboard/summary_annonces.html'

    def __init__(self, request, user_only):
        super().__init__(request)
        self.user_only = user_only

    def get_context_data(self, parent_context):
        annonces = Annonce.objects.all()
        if self.user_only:
            annonces = annonces.filter(owner=self.request.user)
        return {'total_annonces': annonces.count()}


@hooks.register('construct_homepage_summary_items')
def add_annonces_summary_item(request, items):
    """Ajoute le nombre de petites annonces sur le tableau de bord."""
    user_only = None
    if request.user.has_perm('echanges.edit_annonce'):
        user_only = False
    elif request.user.has_perm('echanges.add_annonce'):
        user_only = True
    if user_only is not None:
        items.append(AnnoncesSummaryItem(request, user_only))


# GESTION DES PETITES ANNONCES
# ------------------------------------------------------------------------------


class AnnonceButtonHelper(ButtonHelper):
    def add_button(self, *args, **kwargs):
        # améliore le texte du bouton d'ajout
        button = super().add_button(*args, **kwargs)
        button['label'] = "Ajouter une petite annonce"
        button['title'] = "Ajouter une nouvelle petite annonce"
        return button


class AnnoncePermissionHelper(PermissionHelper):
    def user_can_list(self, user):
        # affiche la liste et l'élément de menu si l'utilisateur
        # peut ajouter une petite annonce ou les modifier
        return self.user_can_create(user) or self.user_can_edit(user)

    def user_can_edit(self, user):
        perm_codename = self.get_perm_codename('change')
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_inspect_obj(self, user, obj):
        if not self.inspect_view_enabled:
            return False
        perms = obj.permissions_for_user(user)
        return perms.is_owner()

    def user_can_edit_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_edit()

    def user_can_delete_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_delete()


class AnnonceAdmin(ModelAdmin):
    """
    Administration des petites annonces pour la personne connectée.
    """

    model = Annonce
    menu_label = "Petites annonces"
    menu_icon = 'bullhorn'
    # étends les classes utilisées par défaut
    button_helper_class = AnnonceButtonHelper
    permission_helper_class = AnnoncePermissionHelper
    create_view_class = AnnonceCreateView
    index_view_class = AnnonceIndexView
    # personnalise la liste des annonces
    ordering = ['-last_modified_at']
    list_display = (
        'photo',
        'title_formatted',
        'type_formatted',
        'category',
        'updated',
        'is_active',
    )
    list_display_add_buttons = 'title_formatted'
    list_filter = ('type', 'category', 'is_active')
    list_per_page = 15

    def get_queryset(self, request):
        qs = super().get_queryset(request).select_related('owner')
        if self.is_user_annonces_only(request):
            # filtre avec les annonces de la personne
            qs = qs.filter(owner=request.user)
        return qs

    def get_list_display(self, request):
        list_display = self.list_display
        if not self.is_user_annonces_only(request):
            # affiche également l'annonceur-se
            list_display += ('owner_formatted',)
        return list_display

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'title_formatted':
            return ['title']
        return []

    def is_user_annonces_only(self, request):
        """Détermine s'il ne faut afficher que les annonces de la personne."""
        return (
            AnnonceIndexView.USER_ANNONCES_VAR in request.GET
            or not request.user.has_perm('echanges.change_annonce')
        )

    # Mise en forme des champs ------------------------------------------------

    def photo(self, obj):
        photo = obj.get_highlighted_photo()
        if not photo:
            return ''
        return format_html(
            '<img src="{}" class="photo-thumb" alt="{}">',
            photo.photo.thumbnail.url,
            photo.title,
        )

    photo.short_description = "photo"

    def title_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette petite annonce">{}</a></h2>',
            obj.get_absolute_url(),
            obj.title,
        )

    title_formatted.short_description = "titre"
    title_formatted.admin_order_field = 'title'

    def type_formatted(self, obj):
        return format_html(
            '<span class="tag no-icon">{}</span>', obj.get_type_display()
        )

    type_formatted.short_description = "type"
    type_formatted.admin_order_field = 'type'

    def updated(self, obj):
        return postdatetime(obj.last_modified_at)

    updated.short_description = "mise à jour"
    updated.admin_order_field = 'last_modified_at'

    def owner_formatted(self, obj):
        return format_html(
            '<a href="{}" title="Voir les détails du compte">{}</a>',
            obj.owner.get_admin_details_url(),
            obj.owner.get_short_name() or obj.owner.email,
        )

    owner_formatted.short_description = "annonceur-se"
    owner_formatted.admin_order_field = 'owner'
