from django import forms
from django.core.validators import FileExtensionValidator
from django.forms.models import inlineformset_factory

from modelcluster.forms import ClusterForm
from tapeforms.contrib.foundation import FoundationTapeformMixin
from tinymce.widgets import TinyMCE

from .models import OffreEmploi, OffreEmploiCategory, OffreEmploiPhoto


class OffreEmploisFiltersForm(FoundationTapeformMixin, forms.Form):
    """
    Formulaire de filtres sur les offres d'emplois.
    """

    category = forms.ModelChoiceField(
        queryset=OffreEmploiCategory.objects,
        empty_label="Toutes les catégories",
        required=False,
    )
    query = forms.CharField(
        required=False,
        widget=forms.TextInput(attrs={'placeholder': "Mots clés…"}),
    )

    field_template = 'tapeforms/fields/simple.html'

    @property
    def has_filters(self):
        """
        Détermine si le formulaire a des critères de définis.
        """
        return self.is_bound and (
            self.data.get('category') or self.data.get('query')
        )

    def filter_queryset(self, queryset):
        """Filtre la sélection avec les critères du formulaire.

        Retourne une nouvelle sélection de offres d'emplois à partir de
        `queryset` en la filtrant avec les valeurs du formulaire, après
        l'avoir validé. Si le critère `query` est définis, la sélection
        retournée sera une instance ``SearchResults`` de Wagtail au lieu
        d'une ``QuerySet``.
        """
        category = self.cleaned_data.get('category')
        if category:
            queryset = queryset.filter(category=category)
        query = self.cleaned_data.get('query')
        if query:
            queryset = queryset.search(query, order_by_relevance=False)
        return queryset


class OffreEmploiContactForm(FoundationTapeformMixin, forms.Form):
    """
    Formulaire de contact sur une offre d'emploi.
    """

    message = forms.CharField(
        min_length=5,
        error_messages={
            'required': "Le contenu du message est requis.",
            'min_length': "Le contenu du message est trop court.",
        },
        label="Votre message",
        widget=forms.Textarea,
    )

    document = forms.FileField(
        required=False,
        label="Document à joindre",
        help_text="Format(s) autorisé(s): PDF",
        validators=[
            FileExtensionValidator(
                ['pdf'],
                (
                    "Mauvais format de fichier. Vérifiez le(s) "
                    "format(s) autorisé(s)"
                ),
            ),
        ],
    )


class OffreEmploiPhotoForm(ClusterForm):
    class Meta:
        model = OffreEmploiPhoto
        fields = [
            'photo',
        ]


OffreEmploiPhotoFormSet = inlineformset_factory(
    OffreEmploi,
    OffreEmploiPhoto,
    form=OffreEmploiPhotoForm,
    extra=1,
    can_delete=False,
    can_order=False,
)


class OffreEmploiCreateForm(ClusterForm):
    description = forms.CharField(widget=TinyMCE())

    def __init__(
        self, data=None, files=None, instance=None, prefix=None, **kwargs
    ):
        super().__init__(
            data, files, instance=instance, prefix=prefix, **kwargs
        )
        self.formsets['photos'] = OffreEmploiPhotoFormSet(
            data, files, instance=instance
        )
        self.fields['description'].initial = self.instance._meta.get_field(
            'description'
        ).default

    class Meta:
        model = OffreEmploi
        fields = [
            'title',
            'type',
            'category',
            'description',
            'rue',
            'code_postal',
            'ville',
        ]
        formsets = [
            'photos',
        ]
