import logging
import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.text import slugify

from djchoices import ChoiceItem, DjangoChoices
from dynamic_filenames import FilePattern
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from stdimage.models import StdImageField
from tinymce.widgets import TinyMCE
from wagtail.admin.edit_handlers import (
    FieldPanel,
    HelpPanel,
    InlinePanel,
    MultiFieldPanel,
    ObjectList,
)
from wagtail.contrib.settings.models import BaseSiteSetting, register_setting
from wagtail.core.fields import RichTextField
from wagtail.core.models import Orderable
from wagtail.search import index
from wagtail.search.backends import get_search_backend

from boiteaasso.accounts.models import UserAccount
from boiteaasso.base.models import OptionalAdresseMixin, OwnerLocalisationMixin
from boiteaasso.base.widgets import AdminImageWidget
from boiteaasso.utils.functional import cached_class_property

logger = logging.getLogger(__name__)

"""Le nombre maximum de photos sur une offre d'emploi."""
OFFRE_EMPLOI_PHOTOS_MAX = 3

"""La taille de la vignette d'une photo d'une offre d'emploi."""
OFFRE_EMPLOI_THUMB_SIZE = 150


class OffreEmploiCategory(models.Model):
    """
    Une catégorie utilisée pour qualifier une offre d'emploi.
    """

    name = models.CharField("nom", max_length=255)
    slug = models.SlugField(unique=True, max_length=80)

    class Meta:
        ordering = ['name']
        verbose_name = "catégorie"
        verbose_name_plural = "catégories des offres d'emplois"

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # génère le slug depuis le nom
        self.slug = slugify(self.name)
        return super().save(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [FieldPanel('name', classname='title')]


class OffreEmploiType(DjangoChoices):
    """
    Les types de offres d'emplois disponibles.
    """

    offre = ChoiceItem('offre', "Offre", label_plural="Offres")
    demande = ChoiceItem('demande', "Demande", label_plural="Demandes")


class OffreEmploiQuerySet(models.QuerySet):
    def active(self):
        """Filtre avec les offres d'emplois visibles publiquement."""
        return self.filter(is_active=True)

    def search(self, query, **kwargs):
        """Recherche une offre d'emploi avec des mots clés."""
        backend = get_search_backend()
        kwargs.setdefault('fields', ['title', 'description'])
        return backend.search(query, self, **kwargs)


class OffreEmploi(
    ClusterableModel,
    OwnerLocalisationMixin,
    OptionalAdresseMixin,
    index.Indexed,
):
    """
    Une offre d'emploi, que ce soit une offre ou une demande.
    """

    title = models.CharField("titre", max_length=255)
    type = models.CharField(
        "type", max_length=10, choices=OffreEmploiType.choices
    )
    category = models.ForeignKey(
        OffreEmploiCategory,
        on_delete=models.SET_NULL,
        null=True,
        verbose_name="catégorie",
    )
    is_active = models.BooleanField(
        "Publication confirmée",
        default=False,
    )
    description = models.TextField(
        default="""
        Pensez à présenter ou préciser :
        <ul>
        <li>l'employeur, son univers,</li>
        <li>le travail proposé, le type de contrat, les qualités demandées,</li>
        <li>les précisions sur l'organisation (lieux, horaires de travail).</li>
        </ul>
        """,
    )

    # Localisation
    use_owner_geolocalisation = models.BooleanField(
        "localisation de l'annonceur⋅se de l'offre d'emploi",
        default=False,
        help_text=(
            "Détermine s'il faut utiliser la localisation de l'annonceur⋅se "
            "à la place, si elle est définie."
        ),
    )

    # Méta-données
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='offres_emploi',
        editable=True,
        verbose_name="annonceur⋅se",
    )
    created_at = models.DateTimeField(
        "date de création", default=timezone.now, blank=True, editable=False
    )
    last_modified_at = models.DateTimeField(
        "date de la dernière modification", blank=True, editable=False
    )

    objects = OffreEmploiQuerySet.as_manager()

    class Meta:
        verbose_name = "offre d'emploi"
        verbose_name_plural = "offres d'emplois"
        permissions = (
            (
                'contact_offreemploi',
                "Peut contacter les annonceurs⋅ses",
            ),
        )

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if self.last_modified_at is None or self.pk is not None:
            self.last_modified_at = timezone.now()
        return super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse('emplois:offre_emploi-detail', args=[self.pk])

    def permissions_for_user(self, user):
        return OffreEmploiPermissionTester(self, user)

    # Gestion des médias associés ---------------------------------------------

    @property
    def has_photos(self):
        """
        Détermine si des photos sont définies pour cette offre d'emploi.
        """
        return self.photos.count()

    def get_highlighted_photo(self):
        """
        Récupère, si elle existe, la photo mise en avant pour cette offre
        d'emploi. C'est actuellement la première photo qui est retournée.
        """
        return self.photos.first()

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        MultiFieldPanel(
            [
                FieldPanel('title'),
                FieldPanel('type'),
                FieldPanel('category'),
            ],
            heading="Offre d'emploi",
        ),
        MultiFieldPanel(
            [
                FieldPanel('description', widget=TinyMCE()),
                InlinePanel(
                    'photos',
                    max_num=OFFRE_EMPLOI_PHOTOS_MAX,
                    label="Photos",
                    help_text=(
                        "Vous pouvez ajouter {0} photos, la première sera "
                        "mise en avant et utilisée pour présenter l'offre "
                        "d'emploi.".format(
                            "plusieurs"
                            if not OFFRE_EMPLOI_PHOTOS_MAX
                            else "jusqu'à {}".format(OFFRE_EMPLOI_PHOTOS_MAX)
                        )
                    ),
                ),
            ],
            heading="Description",
        ),
        MultiFieldPanel(
            [FieldPanel('use_owner_geolocalisation')]
            + OptionalAdresseMixin.adresse_panels,
            heading="Localisation",
        ),
    ]

    search_fields = [
        index.SearchField('title', partial_match=True, boost=2),
        index.FilterField('type'),
        index.FilterField('category'),
        index.SearchField('description'),
        index.FilterField('is_active'),
        index.FilterField('owner'),
        index.FilterField('created_at'),
        index.FilterField('last_modified_at'),
    ]


class OffreEmploiPermissionTester:
    """
    Testeur des permissions d'un utilisateur sur une offre d'emploi.
    """

    perm_re = re.compile(r'emplois\.(\w+)_offreemploi')

    def __init__(self, offre_emploi, user):
        self.offre_emploi = offre_emploi
        self.user = user

    @cached_property
    def permissions(self):
        """
        Retourne les permissions de l'utilisateur liées aux offres d'emploi.
        """
        permissions = set()
        for perm in self.user.get_all_permissions():
            match = self.perm_re.match(perm)
            if match:
                permissions.add(match.group(1))
        return permissions

    def is_owner(self):
        return self.offre_emploi.owner_id == self.user.pk

    def can_edit(self):
        return (
            self.user.is_superuser
            or 'change' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_delete(self):
        return (
            self.user.is_superuser
            or 'delete' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_contact(self):
        return not self.is_owner() and 'contact' in self.permissions


class OffreEmploiPhoto(Orderable):
    """
    Une photo au sein d'une offre d'emploi.
    """

    offre_emploi = ParentalKey(
        OffreEmploi, on_delete=models.CASCADE, related_name='photos'
    )
    photo_upload_to = FilePattern(
        filename_pattern='offres_emploi/photos/{uuid:x}{ext}'
    )
    photo = StdImageField(
        upload_to=photo_upload_to,
        variations={
            'thumbnail': {
                'width': OFFRE_EMPLOI_THUMB_SIZE,
                'height': OFFRE_EMPLOI_THUMB_SIZE,
            }
        },
    )
    title = models.CharField("titre", max_length=255, blank=True)

    class Meta:
        ordering = ('offre_emploi', 'sort_order')
        verbose_name = "photo d'une offre d'emploi"
        verbose_name_plural = "photos d'une offre d'emploi"

    def __str__(self):
        return self.title or 'Untitled'

    def clean(self, *args, **kwargs):
        if not self.title:
            # définis le titre depuis le nom du fichier uniquement
            # pour une nouvelle photo (après, le nom est modifié)
            if self.photo.name and not self.photo._committed:
                self.title = self.photo.name.rsplit('.', 1)[0]
            else:
                raise ValidationError(
                    {'title': "Veuillez renseigner un titre pour cette photo."}
                )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        FieldPanel('photo', widget=AdminImageWidget),
        FieldPanel('title'),
    ]


class UserOffreEmploiStats(models.Model):
    """
    Statistiques sur les offres d'emplois pour un utilisateur⋅rice.
    """

    account = models.OneToOneField(
        UserAccount,
        on_delete=models.CASCADE,
        related_name='offre_emploi_stats',
        primary_key=True,
    )
    offre_emplois_nb = models.PositiveSmallIntegerField(
        "nombre d'offres d'emploi",
        default=0,
        help_text="Nombre total de offres d'emplois publiées.",
    )
    messages_nb = models.PositiveSmallIntegerField(
        "nombre de messages",
        default=0,
        help_text=(
            "Nombre total de messages envoyés sur les offres d'emplois."
        ),
    )

    @classmethod
    def increment(cls, counter, account):
        """Incrémente un compteur pour un utilisateur⋅rice."""
        obj, created = cls.objects.get_or_create(
            account=account, defaults={counter: 1}
        )
        if not created:
            setattr(obj, counter, getattr(obj, counter) + 1)
            obj.save()
        return obj


@register_setting(icon='bullhorn')
class EmploisSettings(BaseSiteSetting):
    """
    Paramètres du service d'émplois.
    """

    introduction = models.TextField(
        help_text=(
            "Un petit texte d'introduction pour décrire le service, affiché "
            "en haut de la page."
        )
    )
    last_offre_emplois_nb = models.PositiveSmallIntegerField(
        "dernières offres d'emplois",
        default=3,
        help_text=(
            "Le nombre de dernières offres d'emplois à afficher, ou 0 pour "
            "les masquer."
        ),
    )
    help_is_active = models.BooleanField(
        "aide active",
        default=False,
        help_text="Définis si la page d'aide est visible.",
    )
    help_content = RichTextField("contenu de l'aide", blank=True)

    class Meta:
        verbose_name = "paramètres du service d'émplois"

    def __str__(self):
        return self.site.site_name or "Site par défaut"

    def clean(self, *args, **kwargs):
        if self.help_is_active and not self.help_content:
            raise ValidationError(
                {
                    'help_content': (
                        "Veuillez définir le contenu de la page ou la "
                        "désactiver."
                    )
                }
            )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    @cached_class_property
    def edit_handler(cls):
        category_url = 'emplois_offreemploicategory_modeladmin_index'
        return ObjectList(
            [
                MultiFieldPanel(
                    [
                        HelpPanel(
                            content=format_html(
                                '<a href="{}" class="button">{}</a>',
                                reverse(category_url),
                                "Modifier les catégories",
                            ),
                            template=(
                                'wagtailadmin/edit_handlers/raw_panel.html'
                            ),
                        )
                    ],
                    heading="Offres d'emploi",
                ),
                MultiFieldPanel(
                    [
                        FieldPanel('introduction'),
                        FieldPanel('last_offre_emplois_nb'),
                    ],
                    heading="Page d'accueil",
                ),
                MultiFieldPanel(
                    [FieldPanel('help_is_active'), FieldPanel('help_content')],
                    heading="Page d'aide",
                ),
            ]
        )
