from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import AccessMixin, LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.forms import Form
from django.http import Http404
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.views.generic import (
    CreateView,
    DetailView,
    FormView,
    ListView,
    TemplateView,
)
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import BaseFormView, FormMixin

from wagtail.contrib.modeladmin import views as modeladmin_views
from wagtail.models import Site

from boiteaasso.base.mixins import ConfirmationEmailMixin, MapViewMixin
from boiteaasso.utils.views import CollectionPageMixin

from .forms import (
    OffreEmploiContactForm,
    OffreEmploiCreateForm,
    OffreEmploisFiltersForm,
)
from .models import (
    EmploisSettings,
    OffreEmploi,
    OffreEmploiType,
    UserOffreEmploiStats,
)
from .signals import offreemploi_created


class SettingsMixin:
    @cached_property
    def settings(self):
        """Retourne les paramètres du service d'emplois."""
        return EmploisSettings.for_site(Site.find_for_request(self.request))

    @property
    def help_url(self):
        return (
            None
            if not self.settings.help_is_active
            else self.request.build_absolute_uri(reverse('emplois:help'))
        )


class EmploisPageMixin(CollectionPageMixin, SettingsMixin):
    collection_list_title = "Offres et demandes d'emplois"
    collection_list_url = reverse_lazy('emplois:offres_emplois')
    show_breadcrumb = False

    def collection_include_list_crumb(self):
        return issubclass(self.__class__, DetailView)

    def get_breadcrumb(self):
        return [
            {'title': "Services"},
            {'title': "Emplois"},
        ] + super().get_breadcrumb()


class HelpView(EmploisPageMixin, TemplateView):
    template_name = 'emplois/help.html'
    page_title = "Aide pour les emplois"
    breadcrumb_title = "Aide"

    def get(self, request, *args, **kwargs):
        if not self.settings.help_is_active:
            return redirect(self.collection_list_url)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['content'] = self.settings.help_content
        return context

    def get_page_actions(self):
        return [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les offres d'emplois",
                'label': self.collection_list_title,
            }
        ]


class OffreEmploiListView(EmploisPageMixin, MapViewMixin, ListView):
    """
    Liste les offres d'emplois, éventuellement filtrées.
    """

    context_object_name = 'offres_emplois'
    template_name = 'emplois/offres_emplois_list.html'
    page_seo_title = "Liste des offres d'emplois"
    page_title = "Offres et demandes d'emplois"

    def get_queryset(self):
        """Filtre la sélection de offres d'emplois à afficher."""
        queryset = OffreEmploi.objects.active().order_by('-last_modified_at')
        # filtre avec le type définis dans l'URL
        self.offres_emplois_type = self.kwargs.get('offres_emplois_type', None)
        if self.offres_emplois_type is not None:
            try:
                self.offres_emplois_type = OffreEmploiType.get_choice(
                    self.offres_emplois_type
                )
            except KeyError:
                raise Http404(
                    "Le type de offres d'emplois demandé est inconnu."
                )
            queryset = queryset.filter(type=self.offres_emplois_type.value)
        # filtre avec les autres critères du formulaire
        self.filters_form = OffreEmploisFiltersForm(self.request.GET)
        if self.filters_form.has_filters:
            self.filters_form.full_clean()
            queryset = self.filters_form.filter_queryset(queryset)
        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['offres_emplois_type'] = self.offres_emplois_type
        context['filters_form'] = self.filters_form
        context['introduction'] = self.settings.introduction
        return context

    def get_page_actions(self):
        actions = []
        if self.request.user.has_perm('emplois.add_offreemploi'):
            actions.append(
                {
                    'url': reverse('emplois:offre_emploi-create'),
                    'class': 'icon fa-plus',
                    'title': "Ajouter une annonce",
                    'label': "Ajouter une annonce",
                }
            )
            actions.append(
                {
                    'url': '{}?{}'.format(
                        reverse('emplois_offreemploi_modeladmin_index'),
                        OffreEmploiIndexView.USER_OFFRE_EMPLOIS_VAR,
                    ),
                    'class': 'icon fa-bullhorn',
                    'title': "Gérer mes annonces",
                    'label': "Mes annonces",
                }
            )
        if self.settings.help_is_active:
            actions.append(
                {
                    'url': reverse('emplois:help'),
                    'title': HelpView.page_title,
                    'label': mark_safe(
                        '<i class="fa fa-question-circle" aria-hidden="true">'
                        '</i>'
                    ),
                    'attrs': {'aria-label': "Aide sur les emplois"},
                }
            )
        return actions

    def get_map_options(self, context):
        return {
            'markers': self.get_offres_emplois_markers(context['object_list']),
            'mapOptions': {'center': settings.DEFAULT_MAP_CENTER, 'zoom': 8},
        }

    def get_offres_emplois_markers(self, object_list):
        return [
            {
                'lat': offre_emploi.latitude,
                'lng': offre_emploi.longitude,
                'popup': render_to_string(
                    'emplois/offre_emploi_popup.html',
                    {'offre_emploi': offre_emploi},
                ),
            }
            for offre_emploi in object_list
            if offre_emploi.is_geolocated
        ]


class ContactFormMixin(FormMixin):
    form_class = OffreEmploiContactForm


class OffreEmploiDetailView(
    EmploisPageMixin, MapViewMixin, ContactFormMixin, DetailView
):
    """
    Présente une offre d'emploi.
    """

    context_object_name = 'offre_emploi'
    queryset = OffreEmploi.objects.active()
    template_name = 'emplois/offre_emploi.html'

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)
        # récupère les permissions de l'utilisateur sur l'offre d'emploi
        self.user_perms = obj.permissions_for_user(self.request.user)
        return obj

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['can_contact'] = self.user_perms.can_contact()
        context['form'] = self.get_form()
        return context

    def get_page_title(self):
        return self.object.title

    def get_page_seo_title(self):
        return '{} : {}'.format(
            self.object.get_type_display(), self.object.title
        )

    def get_page_actions(self):
        actions = [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les offres d'emplois",
                'label': self.collection_list_title,
            }
        ]
        if self.user_perms.can_edit():
            actions.append(
                {
                    'url': reverse(
                        'emplois_offreemploi_modeladmin_edit',
                        args=[self.object.pk],
                    ),
                    'class': 'icon fa-pencil',
                    'title': "Modifier cette offre d'emploi",
                    'label': "Modifier…",
                }
            )
        return actions

    def get_map_options(self, context):
        if not self.object.is_geolocated:
            return {}
        return {
            'markers': [
                {
                    'lat': self.object.latitude,
                    'lng': self.object.longitude,
                    'popup': self.object.get_adresse_display(),
                }
            ],
            'mapOptions': {
                'center': [self.object.latitude, self.object.longitude],
                'zoom': 8,
            },
        }


# VUES RESTREINTES
# ------------------------------------------------------------------------------


class OffreEmploiContactView(
    AccessMixin,
    ConfirmationEmailMixin,
    OffreEmploiDetailView,
    SettingsMixin,
    BaseFormView,
):
    """
    Contact l'annonceur⋅se d'une offre ou demande d'emploi.
    """

    email_template = 'emplois/emails/offre_emploi_contact.txt'
    email_template_html = 'emplois/emails/offre_emploi_contact.html'
    email_template_subject = 'emplois/emails/offre_emploi_contact_subject.txt'

    def has_permission(self):
        return self.object.permissions_for_user(self.request.user).can_contact()

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_permission():
            if not request.user.is_authenticated:
                return self.handle_no_permission()
            messages.error(
                self.request,
                "Vous n'avez pas la permission de contacter cette offre.",
            )
            return redirect(self.object)
        return super().dispatch(request, *args, **kwargs)

    # Traitement du formulaire ------------------------------------------------

    def form_valid(self, form):
        # construit le contexte du courriel et l'envoi
        context = {
            'from': self.request.user,
            'message': form.cleaned_data['message'],
        }
        attachment = None
        if form.cleaned_data['document']:
            filename = form.cleaned_data['document'].name
            content = form.cleaned_data['document'].read()
            mimetype = form.cleaned_data['document'].content_type
            attachment = [(filename, content, mimetype)]
        self.send_confirmation_email(
            self.object.owner,
            context,
            reply_to=[self.request.user.email],
            attachments=attachment,
        )

        # incrémente le compteur de la personne
        UserOffreEmploiStats.increment('messages_nb', self.request.user)

        messages.success(
            self.request,
            (
                "Votre message pour {} a été envoyé.".format(
                    self.object.owner.get_full_name() or "l'annonceur⋅se"
                )
            ),
        )

        return redirect(self.object)

    def form_invalid(self, form):
        errors = (
            set().union(*form.errors.values()).union(form.non_field_errors())
        )
        if not errors:
            message = "Impossible d'envoyer votre message."
        elif len(errors) == 1:
            message = "Impossible d'envoyer votre message : {}".format(
                errors.pop().lower()
            )
        else:
            message = "Impossible d'envoyer votre message :{}".format(
                '<br>&nbsp;- '.join(['', *errors])
            )
        messages.error(self.request, message)

        return super().form_invalid(form)


class OffreEmploiCreateView(modeladmin_views.CreateView):
    def get_instance(self):
        # définis l'annonceur⋅se avec la personne connectée
        instance = super().get_instance()
        instance.owner = self.request.user
        return instance

    def form_valid(self, form):
        response = super().form_valid(form)
        offreemploi_created.send(
            sender=self.__class__, instance=form.instance, request=self.request
        )
        return response


class OffreEmploiIndexView(modeladmin_views.IndexView):
    USER_OFFRE_EMPLOIS_VAR = 'self'
    # ajoute un argument pour filtrer avec les offres de la personne
    IGNORED_PARAMS = modeladmin_views.IndexView.IGNORED_PARAMS + (
        USER_OFFRE_EMPLOIS_VAR,
    )


# MODERATION DES ANNONCES D'EMPLOI
# -----------------------------------------------------------------------------


class OffreEmploiModerationMixin(
    AccessMixin,
    SingleObjectMixin,
    SettingsMixin,
    FormView,
    ConfirmationEmailMixin,
):
    model = OffreEmploi
    form_class = Form
    success_url = reverse_lazy('wagtailadmin_home')

    def has_permission(self):
        return self.request.user.is_superuser

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_permission():
            if not request.user.is_authenticated:
                return self.handle_no_permission()
            messages.error(
                self.request, "Vous n'avez pas la permission pour cette action"
            )
            return redirect('wagtaimadmin_home')
        return super().dispatch(request, *args, **kwargs)


class OffreEmploiApproveView(OffreEmploiModerationMixin):
    email_template = 'emplois/emails/offre_emploi_approved.txt'
    email_template_html = 'emplois/emails/offre_emploi_approved.html'
    email_template_subject = 'emplois/emails/offre_emploi_approved_subject.txt'

    def form_valid(self, form):
        self.object.is_active = True
        self.object.save()
        self.send_confirmation_email(self.object.owner)
        m = "Annonce publiée. Un courriel a été expédié à l'auteur."
        messages.success(self.request, m)
        return super().form_valid(form)


class OffreEmploiRejectView(OffreEmploiModerationMixin):
    email_template = 'emplois/emails/offre_emploi_rejected.txt'
    email_template_html = 'emplois/emails/offre_emploi_rejected.html'
    email_template_subject = 'emplois/emails/offre_emploi_rejected_subject.txt'

    def form_valid(self, form):
        self.send_confirmation_email(self.object.owner)
        m = "Annonce rejetée. Un courriel a été expédié à l'auteur."
        messages.warning(self.request, m)
        self.object.delete()
        return super().form_valid(form)


# VUES FRONT
# -----------------------------------------------------------------------------


class OffreEmploiFrontCreateView(
    LoginRequiredMixin,
    EmploisPageMixin,
    CreateView,
    SuccessMessageMixin,
):
    form_class = OffreEmploiCreateForm
    template_name = "emplois/offreemploi_form.html"
    page_title = "Ajouter une annonce"
    success_url = reverse_lazy('emplois:offres_emplois')
    success_message = (
        "Votre annonce a bien été enregistrée et est en attente "
        "de modération."
    )

    def form_valid(self, form):
        form.instance.owner = self.request.user
        response = super().form_valid(form)
        offreemploi_created.send(
            sender=self.__class__, instance=form.instance, request=self.request
        )
        return response
