from django.template.loader import render_to_string
from django.utils.html import format_html

from wagtail.admin.panels import Panel
from wagtail.admin.site_summary import SummaryItem
from wagtail.contrib.modeladmin.helpers import ButtonHelper, PermissionHelper
from wagtail.contrib.modeladmin.options import ModelAdmin
from wagtail.core import hooks

from boiteaasso.base.templatetags.horodated import postdatetime

from .models import OffreEmploi
from .views import OffreEmploiCreateView, OffreEmploiIndexView


class OffreEmploisSummaryItem(SummaryItem):
    """
    Affiche le nombre de offres d'emplois de la personne ou au total.
    """

    order = 900
    template_name = 'emplois/dashboard/summary_offre_emplois.html'

    def __init__(self, request, user_only):
        super().__init__(request)
        self.user_only = user_only

    def get_context_data(self, parent_context):
        offres_emploi = OffreEmploi.objects.all()
        if self.user_only:
            offres_emploi = offres_emploi.filter(owner=self.request.user)
        return {'total_offre_emplois': offres_emploi.count()}


@hooks.register('construct_homepage_summary_items')
def add_offre_emplois_summary_item(request, items):
    """Ajoute le nombre de offres d'emplois sur le tableau de bord."""
    user_only = None
    if request.user.has_perm('emplois.edit_offreemploi'):
        user_only = False
    elif request.user.has_perm('emplois.add_offreemploi'):
        user_only = True
    if user_only is not None:
        items.append(OffreEmploisSummaryItem(request, user_only))


# GESTION DES OFFRES d'EMPLOI
# ------------------------------------------------------------------------------


class OffreEmploiButtonHelper(ButtonHelper):
    def add_button(self, *args, **kwargs):
        # améliore le texte du bouton d'ajout
        button = super().add_button(*args, **kwargs)
        button['label'] = "Ajouter une offre d'emploi"
        button['title'] = "Ajouter une nouvelle offre d'emploi"
        return button


class OffreEmploiPermissionHelper(PermissionHelper):
    def user_can_list(self, user):
        # affiche la liste et l'élément de menu si l'utilisateur
        # peut ajouter une offre d'emploi ou les modifier
        return self.user_can_create(user) or self.user_can_edit(user)

    def user_can_edit(self, user):
        perm_codename = self.get_perm_codename('change')
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_inspect_obj(self, user, obj):
        if not self.inspect_view_enabled:
            return False
        perms = obj.permissions_for_user(user)
        return perms.is_owner()

    def user_can_edit_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_edit()

    def user_can_delete_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_delete()


class OffreEmploiAdmin(ModelAdmin):
    """
    Administration des offres d'emplois pour la personne connectée.
    """

    model = OffreEmploi
    menu_label = "Offres et demandes d'emplois"
    menu_icon = 'bullhorn'
    # étends les classes utilisées par défaut
    button_helper_class = OffreEmploiButtonHelper
    permission_helper_class = OffreEmploiPermissionHelper
    create_view_class = OffreEmploiCreateView
    index_view_class = OffreEmploiIndexView
    # personnalise la liste des offres d'emploi
    ordering = ['-last_modified_at']
    list_display = (
        'photo',
        'title_formatted',
        'type_formatted',
        'category',
        'updated',
        'is_active',
    )
    list_display_add_buttons = 'title_formatted'
    list_filter = ('type', 'category', 'is_active')
    list_per_page = 15

    def get_queryset(self, request):
        qs = super().get_queryset(request).select_related('owner')
        if self.is_user_offre_emplois_only(request):
            # filtre avec les offres d'emploi de la personne
            qs = qs.filter(owner=request.user)
        return qs

    def get_list_display(self, request):
        list_display = self.list_display
        if not self.is_user_offre_emplois_only(request):
            # affiche également l'annonceur⋅se
            list_display += ('owner_formatted',)
        return list_display

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'title_formatted':
            return ['title']
        return []

    def is_user_offre_emplois_only(self, request):
        """
        Détermine s'il ne faut afficher que les offre d'emploi de la personne.
        """
        return (
            OffreEmploiIndexView.USER_OFFRE_EMPLOIS_VAR in request.GET
            or not request.user.has_perm('emplois.change_offreemploi')
        )

    # Mise en forme des champs ------------------------------------------------

    def photo(self, obj):
        photo = obj.get_highlighted_photo()
        if not photo:
            return ''
        return format_html(
            '<img src="{}" class="photo-thumb" alt="{}">',
            photo.photo.thumbnail.url,
            photo.title,
        )

    photo.short_description = "photo"

    def title_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette offre d\'emploi">{}</a></h2>',
            obj.get_absolute_url(),
            obj.title,
        )

    title_formatted.short_description = "titre"
    title_formatted.admin_order_field = 'title'

    def type_formatted(self, obj):
        return format_html(
            '<span class="tag no-icon">{}</span>', obj.get_type_display()
        )

    type_formatted.short_description = "type"
    type_formatted.admin_order_field = 'type'

    def updated(self, obj):
        return postdatetime(obj.last_modified_at)

    updated.short_description = "mise à jour"
    updated.admin_order_field = 'last_modified_at'

    def owner_formatted(self, obj):
        return format_html(
            '<a href="{}" title="Voir les détails du compte">{}</a>',
            obj.owner.get_admin_details_url(),
            obj.owner.get_short_name() or obj.owner.email,
        )

    owner_formatted.short_description = "annonceur⋅se"
    owner_formatted.admin_order_field = 'owner'


class OffresEmploiForModerationPanel(Panel.BoundPanel):
    name = 'offresemploi_for_moderation'
    order = 202

    def __init__(self, request):
        self.request = request
        if request.user.is_superuser:
            self.offresemploi_moderation = OffreEmploi.objects.filter(
                is_active=False,
            ).order_by('-created_at')
        else:
            self.offresemploi_moderation = None

    def render_html(self, parent_context):
        return render_to_string(
            'emplois/dashboard/offresemploi_for_moderation.html',
            {'offresemploi_moderation': self.offresemploi_moderation},
            request=self.request,
        )


@hooks.register('construct_homepage_panels')
def add_offresemploi_moderation_panel(request, panels):
    panels.append(OffresEmploiForModerationPanel(request))
