from django import forms

from tapeforms.contrib.foundation import FoundationTapeformMixin
from tapeforms.fieldsets import TapeformFieldsetsMixin

from boiteaasso.base.forms import AdresseFormMixin

from .models import FormationCategory, Subscription


class FormationsFiltersForm(FoundationTapeformMixin, forms.Form):
    """
    Formulaire de filtres sur les formations.
    """

    category = forms.ModelChoiceField(
        queryset=FormationCategory.objects,
        empty_label="Toutes les catégories",
        required=False,
    )
    query = forms.CharField(
        required=False,
        widget=forms.TextInput(attrs={'placeholder': "Mots clés…"}),
    )

    field_template = 'tapeforms/fields/simple.html'

    @property
    def has_filters(self):
        """
        Détermine si le formulaire a des critères de définis.
        """
        return self.is_bound and (
            self.data.get('category') or self.data.get('query')
        )

    def filter_queryset(self, queryset):
        """Filtre la sélection avec les critères du formulaire.

        Retourne une nouvelle sélection de formations à partir de
        `queryset` en la filtrant avec les valeurs du formulaire, après
        l'avoir validé. Si le critère `query` est définis, la sélection
        retournée sera une instance ``SearchResults`` de Wagtail au lieu
        d'une ``QuerySet``.
        """
        category = self.cleaned_data.get('category')
        if category:
            queryset = queryset.filter(category=category)
        query = self.cleaned_data.get('query')
        if query:
            queryset = queryset.search(query, order_by_relevance=False)
        return queryset


class FormationSubscribeForm(
    FoundationTapeformMixin,
    AdresseFormMixin,
    TapeformFieldsetsMixin,
    forms.ModelForm,
):
    class Meta:
        model = Subscription
        fields = [
            'name',
            'family_name',
            'phone',
            'email',
            'rue',
            'code_postal',
            'ville',
            'association',
            'picture_right',
            'knowledges',
            'wishes',
            'questions',
        ]

    def __init__(self, formation, **kwargs):
        self.formation = formation
        return super().__init__(**kwargs)

    def get_fieldsets(self):
        asked_fields = [
            f
            for f in self.fields
            if getattr(self.formation, 'ask_{}'.format(f), False)
        ]
        if asked_fields:
            optionnal = {
                'fields': asked_fields,
                'extra': {'legend': "Compléments d'inscription à la formation"},
            }
            return super().get_fieldsets(self.fieldsets + [optionnal])
        return super().get_fieldsets()

    fieldsets = [
        {
            'fields': ('name', 'family_name', 'phone', 'email'),
            'extra': {'legend': "Vos coordonnées"},
        },
        {
            'fields': ('rue', 'code_postal', 'ville'),
            'extra': {'legend': "Votre adresse"},
        },
    ]
