import re
from datetime import datetime, timedelta

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.text import slugify

from dynamic_filenames import FilePattern
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from stdimage.models import StdImageField
from wagtail.admin.panels import (
    FieldPanel,
    FieldRowPanel,
    HelpPanel,
    InlinePanel,
    MultiFieldPanel,
    ObjectList,
    TabbedInterface,
)
from wagtail.contrib.settings.models import BaseSiteSetting, register_setting
from wagtail.fields import RichTextField
from wagtail.models import Orderable
from wagtail.search import index
from wagtail.search.backends import get_search_backend

from boiteaasso.base.models import AdresseMixin, OwnerLocalisationMixin
from boiteaasso.base.widgets import AdminImageWidget
from boiteaasso.utils.functional import cached_class_property

from .edit_handlers import SubscriptionsPanel

"""Le nombre maximum d'image pour une formation."""
FORMATION_PICTURES_MAX = 3

"""La taille de la vignette d'une image d'une formation."""
FORMATION_THUMB_SIZE = 150


class FormationCategory(models.Model):
    """
    Une catégorie utilisée pour qualifier une formation.
    """

    name = models.CharField("nom", max_length=255)
    slug = models.SlugField(unique=True, max_length=80)

    # Picture
    icon_upload_to = FilePattern(
        filename_pattern='formations/icons/{uuid:x}{ext}'
    )
    icon = StdImageField(
        "icon", upload_to=icon_upload_to, blank=True, null=True
    )

    class Meta:
        ordering = ['name']
        verbose_name = "catégorie"
        verbose_name_plural = "catégories des formations"

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # génère le slug depuis le nom
        self.slug = slugify(self.name)
        return super().save(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [FieldPanel('name', classname='title'), FieldPanel('icon')]


class FormationQuerySet(models.QuerySet):
    def active(self):
        """Filtre avec les formations visibles publiquement."""
        return self.filter(is_active=True)

    def future(self):
        """Filtre avec les formations en cours ou à venir."""
        return self.filter(end_at__gt=timezone.now()).order_by('begin_at')

    def recent(self):
        """Filtre avec les formations fraichement passées ou à venir."""
        return self.filter(end_at__gt=timezone.now() - timedelta(30, 0))

    def search(self, query, **kwargs):
        """Recherche une formation avec des mots clés."""
        backend = get_search_backend()
        kwargs.setdefault('fields', ['title', 'description'])
        return backend.search(query, self, **kwargs)


class SubscriptionPermissionTester:
    """
    Testeur des permissions d'un utilisateur sur une inscription.
    """

    perm_re = re.compile(r'formations\.(\w+)_subscription')

    def __init__(self, subscription, user):
        self.subscription = subscription
        self.formation = subscription.formation
        self.user = user

    @cached_property
    def permissions(self):
        """Retourne les permissions de l'utilisateur liées aux formations."""
        permissions = set()
        for perm in self.user.get_all_permissions():
            match = self.perm_re.match(perm)
            if match:
                permissions.add(match.group(1))
        return permissions

    def is_owner(self):
        return self.subscription.user_id == self.user.pk

    def is_organiser(self):
        return self.formation.owner_id == self.user.pk

    def can_edit(self):
        return (
            self.user.is_superuser
            or self.is_organiser()
            or 'change' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_delete(self):
        return (
            self.user.is_superuser
            or self.is_organiser()
            or 'delete' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )


class SubscriptionQuerySet(models.QuerySet):
    def recent(self):
        """Filtre avec les formations fraichement passées ou à venir."""
        return self.filter(
            formation__end_at__gt=timezone.now() - timedelta(30, 0)
        )


class Subscription(AdresseMixin, models.Model):
    """
    Inscription à une formation.
    """

    # Manager par défaut

    objects = SubscriptionQuerySet.as_manager()

    class Meta:
        verbose_name = "Inscription à une formation"
        verbose_name_plural = "Inscriptions à des formations"

    formation = ParentalKey(
        'Formation',
        on_delete=models.CASCADE,
        related_name='subscriptions',
        limit_choices_to={'is_active': True},
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='formation_subscriptions',
        verbose_name="Compte d'inscription",
        limit_choices_to={'is_active': True},
        editable=False,
    )
    submit_time = models.DateTimeField(
        verbose_name="date d'inscription", auto_now_add=True, editable=False
    )
    is_active = models.BooleanField("inscription confirmée", default=False)
    name = models.CharField("prénom", max_length=255)
    family_name = models.CharField("nom", max_length=255)
    email = models.EmailField("Courriel")
    phone = models.CharField("téléphone", max_length=255)

    picture_right = models.BooleanField(
        "droit à l'image",
        default=False,
        help_text=(
            "J'autorise l'organisateur de la formation de diffuser "
            "des images ou j'apparait."
        ),
    )
    association = models.CharField(
        "association",
        max_length=255,
        help_text="Si vous faites partie d'une association, indiquez laquelle.",
        blank=True,
        null=True,
    )
    knowledges = models.TextField(
        "connaissances",
        help_text="Avez vous des connaissances liées à la formation ?",
        blank=True,
        null=True,
    )
    wishes = models.TextField(
        "attentes",
        help_text="Quelles attentes avez vous sur la formation ?",
        blank=True,
        null=True,
    )
    questions = models.TextField(
        "questions",
        help_text="Avez vous des questions pour l'organisateur ?",
        blank=True,
        null=True,
    )
    formation_panels = [
        MultiFieldPanel([FieldPanel('is_active')], heading="Validation")
    ]
    identity_panels = [
        MultiFieldPanel(
            [
                FieldPanel('name'),
                FieldPanel('family_name'),
                FieldPanel('email'),
                FieldPanel('phone'),
            ],
            heading="Identité",
        )
    ]
    address_panels = [
        MultiFieldPanel(
            AdresseMixin.adresse_panels[:-1], heading="Adresse postale"
        )
    ]
    misc_panels = [
        MultiFieldPanel(
            [
                FieldPanel('association'),
                FieldPanel('picture_right'),
                FieldPanel('knowledges'),
                FieldPanel('wishes'),
                FieldPanel('questions'),
            ],
            heading="Compléments",
        )
    ]
    panels = formation_panels + identity_panels + address_panels + misc_panels

    def __str__(self):
        return "inscription de {} {} à « {} »".format(
            self.name, self.family_name, self.formation
        )

    def permissions_for_user(self, user):
        return SubscriptionPermissionTester(self, user)


class Formation(
    ClusterableModel, OwnerLocalisationMixin, AdresseMixin, index.Indexed
):
    """
    Une formation, que ce soit une offre ou une demande.
    """

    title = models.CharField("titre", max_length=255)
    goal = models.CharField("objectif", max_length=255)
    category = models.ForeignKey(
        FormationCategory,
        on_delete=models.SET_NULL,
        null=True,
        verbose_name="catégorie",
    )
    is_active = models.BooleanField(
        "active",
        default=True,
        help_text="Détermine si la formation est visible publiquement.",
    )
    description = RichTextField(
        features=['h2', 'h3', 'bold', 'italic', 'link', 'ul']
    )
    remote = models.BooleanField(
        "visioconférence",
        default=False,
        blank=True,
        help_text="À cocher si la formation est accessible en visio.",
    )
    location = models.CharField(
        "lieu", max_length=255, help_text='Nom commun du lieu de formation'
    )
    max_attendees = models.PositiveSmallIntegerField(
        "nombre maximal de participants",
        default=0,
        help_text=(
            "Limite au delà de laquelle les inscrits sont placés en liste "
            "d'attente. Ignoré si nul."
        ),
    )

    # Localisation
    use_owner_geolocalisation = models.BooleanField(
        "localisation de l'organisateur-trice",
        default=False,
        help_text=(
            "Détermine s'il faut utiliser la localisation de "
            "l'organisateur-trice à la place, si elle est définie."
        ),
    )

    # Méta-données
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='formations',
        editable=True,
        verbose_name="organisateur-trice",
    )
    created_at = models.DateTimeField(
        "date de création", default=timezone.now, blank=True, editable=False
    )
    last_modified_at = models.DateTimeField(
        "date de la dernière modification", blank=True, editable=False
    )

    # Affichage des champs de formulaire
    ask_picture_right = models.BooleanField(
        "droit à l'image",
        help_text="Demander le droit à l'image",
        default=False,
    )
    ask_association = models.BooleanField(
        "association",
        help_text="Demander dans quelle association la personne est engagée",
        default=False,
    )
    ask_knowledges = models.BooleanField(
        "connaissances",
        help_text=(
            "Demander si la personne a des connaissance relatives "
            "à la formation"
        ),
        default=False,
    )
    ask_wishes = models.BooleanField(
        "attentes",
        help_text=(
            "Demander si la personne a des attentes relatives à la formation"
        ),
        default=False,
    )
    ask_questions = models.BooleanField(
        "questions",
        help_text=(
            "Demander si la personne a des questions au sujet de la formation"
        ),
        default=False,
    )

    # Début et fin

    """
    On redonde la date de début et de fin de l'évènement pour faire travailler
    la db en select et order. Il semble impossible d'obtenir des bons résultats
    sans ça car le order_by sur des ranges multiples est complexe.
    """

    begin_at = models.DateTimeField(editable=False, null=True)
    end_at = models.DateTimeField(editable=False, null=True)

    # Manager par défaut

    objects = FormationQuerySet.as_manager()

    class Meta:
        permissions = (
            ('subscribe_formation', "Peut s'inscrire aux formations"),
        )
        ordering = ['-begin_at']

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if self.last_modified_at is None or self.pk is not None:
            self.last_modified_at = timezone.now()

        ranges = self.ranges.all().order_by('begin')
        if ranges.exists():
            self.begin_at = ranges.first().begin
            self.end_at = datetime.combine(
                ranges.last().begin, ranges.last().end
            )
            # combine take time's TZ but it may be naïve
            if timezone.is_naive(self.end_at):
                self.end_at = timezone.make_aware(self.end_at)

        return super().save(*args, **kwargs)

    def get_absolute_url(self, kwargs=None):
        return reverse(
            'formations:formation-detail', args=[self.pk], kwargs=kwargs
        )

    def permissions_for_user(self, user):
        return FormationPermissionTester(self, user)

    def is_lacking_places(self):
        return (
            self.max_attendees
            and self.max_attendees - self.subscriptions.all().count() <= 0
        )

    def remaining_places(self):
        if not self.max_attendees:
            return "∞"
        return (
            self.max_attendees
            - self.subscriptions.filter(is_active=True).count()
        )

    # Gestion des médias associés ---------------------------------------------

    @property
    def has_pictures(self):
        """
        Détermine si des images sont définies pour cette petite annone.
        """
        return self.pictures.count()

    def get_highlighted_picture(self):
        """
        Récupère, si elle existe, l'image mise en avant pour cette petite
        annonce. C'est actuellement la première image qui est retournée.
        """
        return self.pictures.first()

    # Propriétés de Wagtail ---------------------------------------------------

    content_panels = [
        MultiFieldPanel(
            [
                FieldPanel('title'),
                FieldPanel('goal'),
                FieldPanel('category'),
                FieldPanel('is_active'),
                FieldPanel('max_attendees'),
            ],
            heading="Formation",
        ),
        MultiFieldPanel(
            [
                InlinePanel(
                    'ranges',
                    label="Plage horaire",
                    min_num=1,
                    panels=[
                        FieldRowPanel([FieldPanel('begin'), FieldPanel('end')])
                    ],
                )
            ],
            "Plage(s) horaire(s)",
        ),
        MultiFieldPanel(
            [
                InlinePanel(
                    'more_description',
                    label="supplément de description",
                    panels=[
                        FieldRowPanel(
                            [FieldPanel('intitule'), FieldPanel('description')]
                        )
                    ],
                )
            ],
            "supplément(s) de description(s)",
        ),
        MultiFieldPanel([FieldPanel('description')], heading="Description"),
        MultiFieldPanel(
            [
                InlinePanel(
                    'pictures',
                    max_num=FORMATION_PICTURES_MAX,
                    label="Images",
                    help_text=(
                        "Vous pouvez ajouter {0} images.".format(
                            "plusieurs"
                            if not FORMATION_PICTURES_MAX
                            else "jusqu'à {}".format(FORMATION_PICTURES_MAX)
                        )
                    ),
                )
            ],
            heading="Images",
        ),
        MultiFieldPanel(
            [
                FieldPanel('remote'),
                FieldPanel('location'),
                FieldPanel('use_owner_geolocalisation'),
            ]
            + AdresseMixin.adresse_panels,
            heading="Localisation",
        ),
    ]

    form_panels = [
        MultiFieldPanel(
            [
                FieldPanel('ask_association'),
                FieldPanel('ask_picture_right'),
                FieldPanel('ask_knowledges'),
                FieldPanel('ask_wishes'),
                FieldPanel('ask_questions'),
            ],
            heading="Champs optionnels du formulaire d'inscription",
        )
    ]

    subscribtion_panels = [SubscriptionsPanel()]

    edit_handler = TabbedInterface(
        [
            ObjectList(content_panels, heading="Détails de la formation"),
            ObjectList(form_panels, heading="Formulaire de réponse"),
            ObjectList(subscribtion_panels, heading="Inscriptions"),
        ]
    )

    search_fields = [
        index.SearchField('title', partial_match=True, boost=2),
        index.SearchField('goal'),
        index.FilterField('category'),
        index.SearchField('description'),
        index.FilterField('is_active'),
        index.FilterField('owner'),
        index.FilterField('created_at'),
        index.FilterField('last_modified_at'),
        index.FilterField('begin_at'),
        index.FilterField('end_at'),
    ]


class FormationPermissionTester:
    """
    Testeur des permissions d'un utilisateur sur une formation.
    """

    perm_re = re.compile(r'formations\.(\w+)_formation')

    def __init__(self, formation, user):
        self.formation = formation
        self.user = user

    @cached_property
    def permissions(self):
        """Retourne les permissions de l'utilisateur liées aux formations."""
        permissions = set()
        for perm in self.user.get_all_permissions():
            match = self.perm_re.match(perm)
            if match:
                permissions.add(match.group(1))
        return permissions

    def is_owner(self):
        return self.formation.owner_id == self.user.pk

    def can_edit(self):
        return (
            self.user.is_superuser
            or 'change' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_delete(self):
        return (
            self.user.is_superuser
            or 'delete' in self.permissions
            or ('add' in self.permissions and self.is_owner())
        )

    def can_subscribe(self):
        return 'subscribe' in self.permissions


class FormationDescription(Orderable):
    """
    Un élément de description pour une formation.
    """

    intitule = models.CharField("intitulé", max_length=255)
    description = models.CharField("description", max_length=255)
    formation = ParentalKey(
        Formation, on_delete=models.CASCADE, related_name='more_description'
    )

    def copy(self, formation):
        return FormationDescription(
            intitule=self.intitule,
            description=self.description,
            formation=formation,
        )


class FormationTimeRange(models.Model):
    """
    Une plage horaire pour une formation.
    """

    begin = models.DateTimeField('début')
    end = models.TimeField('fin')
    formation = ParentalKey(
        Formation, on_delete=models.CASCADE, related_name='ranges'
    )

    def copy(self, formation):
        return FormationTimeRange(
            begin=self.begin,
            end=self.end,
            formation=formation,
        )

    class Meta:
        ordering = ['begin']


class FormationPicture(Orderable):
    """
    Une image pour une formation.
    """

    formation = ParentalKey(
        Formation, on_delete=models.CASCADE, related_name='pictures'
    )
    picture_upload_to = FilePattern(
        filename_pattern='formations/pictures/{uuid:x}{ext}'
    )
    picture = StdImageField(
        "image",
        upload_to=picture_upload_to,
        variations={
            'thumbnail': {
                'width': FORMATION_THUMB_SIZE,
                'height': FORMATION_THUMB_SIZE,
            }
        },
    )
    title = models.CharField("titre", max_length=255, blank=True)

    class Meta:
        ordering = ('formation', 'sort_order')
        verbose_name = "image d'une formation"
        verbose_name_plural = "images d'une formation"

    def __str__(self):
        return self.title or 'Untitled'

    def copy(self, formation):
        return FormationPicture(
            formation=formation,
            picture=self.picture,
            title=self.title,
        )

    def clean(self, *args, **kwargs):
        if not self.title:
            # définis le titre depuis le nom du fichier uniquement
            # pour une nouvelle picture (après, le nom est modifié)
            if self.picture.name and not self.picture._committed:
                self.title = self.picture.name.rsplit('.', 1)[0]
            else:
                raise ValidationError(
                    {'title': "Veuillez renseigner un titre pour cette image."}
                )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    panels = [
        FieldPanel('picture', widget=AdminImageWidget),
        FieldPanel('title'),
    ]


@register_setting(icon='graduation-cap')
class FormationsSettings(BaseSiteSetting):
    """
    Paramètres du service de formations.
    """

    introduction = models.TextField(
        help_text=(
            "Un petit texte d'introduction pour décrire le service, affiché "
            "en haut de la page."
        )
    )
    last_formations_nb = models.PositiveSmallIntegerField(
        "dernières formations",
        default=3,
        help_text=(
            "Le nombre de dernières formations à afficher, ou 0 pour "
            "les masquer."
        ),
    )
    help_is_active = models.BooleanField(
        "aide active",
        default=False,
        help_text="Définis si la page d'aide est visible.",
    )
    help_content = RichTextField("contenu de l'aide", blank=True)

    class Meta:
        verbose_name = "paramètres du service de formations"

    def __str__(self):
        return self.site.site_name or "Site par défaut"

    def clean(self, *args, **kwargs):
        if self.help_is_active and not self.help_content:
            raise ValidationError(
                {
                    'help_content': (
                        "Veuillez définir le contenu de la page ou la "
                        "désactiver."
                    )
                }
            )
        super().clean(*args, **kwargs)

    # Propriétés de Wagtail ---------------------------------------------------

    @cached_class_property
    def edit_handler(cls):
        category_url = 'formations_formationcategory_modeladmin_index'
        return ObjectList(
            [
                MultiFieldPanel(
                    [
                        HelpPanel(
                            content=format_html(
                                '<a href="{}" class="button">{}</a>',
                                reverse(category_url),
                                "Modifier les catégories",
                            ),
                            template=(
                                'wagtailadmin/edit_handlers/raw_panel.html'
                            ),
                        )
                    ],
                    heading="Formations",
                ),
                MultiFieldPanel(
                    [
                        FieldPanel('introduction'),
                        FieldPanel('last_formations_nb'),
                    ],
                    heading="Page d'accueil",
                ),
                MultiFieldPanel(
                    [FieldPanel('help_is_active'), FieldPanel('help_content')],
                    heading="Page d'aide",
                ),
            ]
        )
