import pytest

from boiteaasso.accounts.tests_accounts.factories import (
    AssociationUserAccountFactory,
    UserAccountFactory,
)

from .factories import FormationFactory, SubscriptionFactory


@pytest.mark.django_db
class TestFormationPermissionTester:
    def setup_method(self):
        self.formation = FormationFactory()

    def test_owner_permissions(self):
        tester = self.formation.permissions_for_user(self.formation.owner)
        assert tester.is_owner()
        assert tester.can_edit()
        assert tester.can_delete()
        assert tester.can_subscribe()

    def test_other_permissions(self):
        tester = self.formation.permissions_for_user(
            AssociationUserAccountFactory()
        )
        assert not tester.is_owner()
        assert not tester.can_edit()
        assert not tester.can_delete()
        assert tester.can_subscribe()

    def test_permitted_user_permissions(self):
        tester = self.formation.permissions_for_user(
            UserAccountFactory(
                permissions__codenames=['add_formation', 'subscribe_formation']
            )
        )
        assert not tester.is_owner()
        assert not tester.can_edit()
        assert not tester.can_delete()
        assert tester.can_subscribe()

    def test_unpermitted_user_permissions(self):
        tester = self.formation.permissions_for_user(UserAccountFactory())
        assert not tester.is_owner()
        assert not tester.can_edit()
        assert not tester.can_delete()
        assert not tester.can_subscribe()

    def test_admin_permissions(self):
        tester = self.formation.permissions_for_user(
            UserAccountFactory(
                permissions__codenames=[
                    'add_formation',
                    'change_formation',
                    'delete_formation',
                    'subscribe_formation',
                ]
            )
        )
        assert not tester.is_owner()
        assert tester.can_edit()
        assert tester.can_delete()
        assert tester.can_subscribe()


@pytest.mark.django_db
class TestSubscriptionPermissionTester:
    def setup_method(self):
        self.subscription = SubscriptionFactory()

    def test_formation_owner_permissions(self):
        tester = self.subscription.permissions_for_user(
            self.subscription.formation.owner
        )
        assert not tester.is_owner()
        assert tester.is_organiser()
        assert tester.can_edit()
        assert tester.can_delete()

    def test_other_association_permissions(self):
        tester = self.subscription.permissions_for_user(
            AssociationUserAccountFactory()
        )
        assert not tester.is_owner()
        assert not tester.is_organiser()
        assert not tester.can_edit()
        assert not tester.can_delete()

    def test_permitted_other_user_permissions(self):
        tester = self.subscription.permissions_for_user(
            UserAccountFactory(permissions__codenames=['add_subscription'])
        )
        assert not tester.is_owner()
        assert not tester.is_organiser()
        assert not tester.can_edit()
        assert not tester.can_delete()

    def test_unpermitted_other_user_permissions(self):
        tester = self.subscription.permissions_for_user(UserAccountFactory())
        assert not tester.is_owner()
        assert not tester.is_organiser()
        assert not tester.can_edit()
        assert not tester.can_delete()

    def test_admin_permissions(self):
        tester = self.subscription.permissions_for_user(
            UserAccountFactory(
                permissions__codenames=[
                    'add_subscription',
                    'change_subscription',
                    'delete_subscription',
                ]
            )
        )
        assert not tester.is_owner()
        assert not tester.is_organiser()
        assert tester.can_edit()
        assert tester.can_delete()


@pytest.mark.django_db
class TestSubscription:
    def setup_method(self):
        self.subscription = SubscriptionFactory(is_active=True)

    def test_str(self):
        assert str(self.subscription)

    def test_remaining_places(self):
        assert self.subscription.formation.remaining_places() == 11

    def test_infinite_remaining_places(self):
        self.subscription.formation.max_attendees = 0
        assert self.subscription.formation.remaining_places() == "∞"
