import datetime

from django.conf import settings
from django.contrib.messages import ERROR as ERROR_LEVEL
from django.contrib.messages import SUCCESS as SUCCESS_LEVEL
from django.contrib.messages import WARNING as WARNING_LEVEL
from django.contrib.messages import get_messages
from django.urls import reverse, reverse_lazy
from django.utils import timezone

import pytest
from bs4 import BeautifulSoup as bs
from wagtail.test.utils.form_data import (
    inline_formset,
    nested_form_data,
    rich_text,
)

from boiteaasso.accounts.tests_accounts.factories import (
    AssociationUserAccountFactory,
    PersonneUserAccountFactory,
)
from boiteaasso.utils.tests import (
    count_text_in_content,
    log_in_client,
    skipif_sqlite,
)

from .factories import (
    FormationCategoryFactory,
    FormationFactory,
    SubscriptionFactory,
)


@pytest.mark.django_db
class TestFormationList:
    url = reverse_lazy('formations:formations')

    def test_get(self, client):
        formations = FormationFactory.create_batch(3)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['formations']) == len(formations)

    def test_get_future_only(self, client):
        formations = FormationFactory.create_batch(4)
        old = FormationFactory()
        old.ranges.update(begin=timezone.now() - datetime.timedelta(30, 0))
        old.save()

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['formations']) == len(formations)

    def test_get_owner(self, client):
        formations = FormationFactory.create_batch(4)
        assert log_in_client(client, formations[0].owner)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['formations']) == len(formations)

    def test_filter_category(self, client):
        category = FormationCategoryFactory()
        formations = [
            FormationFactory(category=category)
        ] + FormationFactory.create_batch(2)

        # la catégorie contient 1 formations
        response = client.get(self.url, {'category': category.pk})
        assert response.status_code == 200
        assert len(response.context['formations']) == 1

        # la catégorie est inexistante, on affiche tout
        response = client.get(self.url, {'category': 'nope'})
        assert response.status_code == 200
        assert len(response.context['formations']) == len(formations)

        # accès par l'onglet
        self.url = reverse_lazy(
            'formations:formations-by-category', args=[category.slug]
        )
        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['formations']) == 1

    @skipif_sqlite
    def test_filter_query(self, client):
        FormationFactory(title="Test azertyuiop")
        FormationFactory.create_batch(2)

        # des mots clés entiers
        response = client.get(self.url, {'query': "Test azertyuiop"})
        assert response.status_code == 200
        assert len(response.context['formations']) == 1

        # un mot clé partiel
        response = client.get(self.url, {'query': "ertyui"})
        assert response.status_code == 200
        assert len(response.context['formations']) == 1

        # un mot clé inexistant
        response = client.get(self.url, {'query': "nope"})
        assert response.status_code == 200
        assert len(response.context['formations']) == 0


@pytest.mark.django_db
class TestFormationDetail:
    def setup_method(self):
        self.formation = FormationFactory()
        self.url = self.formation.get_absolute_url()

    def test_get_anonymous(self, client):
        response = client.get(self.url)
        assert response.status_code == 200
        assert not count_text_in_content(response, "Ajoutée par :")
        assert not count_text_in_content(response, "S'inscrire…")

        actions = response.context['page']['actions']
        assert len(actions) == 1
        assert actions[0]['url'] == reverse('formations:formations')
        assert actions[0]['title'] == "Voir les formations"

    def test_get_owner(self, client):
        assert log_in_client(client, self.formation.owner)
        self.formation.ask_questions = False
        self.formation.save()

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "Ajoutée par :")
        assert count_text_in_content(response, "S'inscrire…")

        actions = response.context['page']['actions']
        # l'organisateur peut modifier sa formation
        assert len(actions) == 3
        assert actions[1]['url'] == reverse(
            'formations_formation_modeladmin_edit', args=[self.formation.pk]
        )
        assert actions[1]['url'].startswith(
            reverse(
                'formations_formation_modeladmin_edit',
                args=[self.formation.pk],
            )
        )
        assert actions[2]['title'] == "Gérer les inscriptions à cette formation"

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_get_other(self, client, logged_user):
        assert log_in_client(client, logged_user())

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "Ajoutée par :")
        assert count_text_in_content(response, "S'inscrire…")
        assert len(response.context['page']['actions']) == 1


@pytest.mark.django_db
class TestFormationDetailLocation:
    def test_no_adresse(self, client):
        formation = FormationFactory()
        assert not formation.has_adresse

        response = client.get(formation.get_absolute_url())
        assert not count_text_in_content(response, "Localisation")

    def test_owner_adresse_not_geolocated(self, client):
        account = AssociationUserAccountFactory(
            association__rue="3 rue des coquelicots",
            association__ville="Libreville",
            association__code_postal="01000",
        )
        formation = FormationFactory(
            owner=account, use_owner_geolocalisation=True
        )
        assert formation.has_adresse
        assert not formation.is_geolocated

        response = client.get(formation.get_absolute_url())
        assert count_text_in_content(response, "Localisation")
        assert count_text_in_content(response, "Libreville")
        assert not count_text_in_content(response, "boiteaassoMapOptions")
        assert not response.context['map_options']

    def test_formation_adresse_geolocated(self, client):
        formation = FormationFactory(
            use_owner_geolocalisation=False,
            rue="3 rue des coquelicots",
            ville="Libreville",
            code_postal="01000",
            geolocalisation="1,2",
        )
        assert formation.has_adresse
        assert formation.is_geolocated

        response = client.get(formation.get_absolute_url())
        assert count_text_in_content(response, "Localisation")
        assert count_text_in_content(response, "Libreville")
        assert count_text_in_content(response, "boiteaassoMapOptions")
        assert response.context['map_options']['markers'] == [
            {
                'lat': '1',
                'lng': '2',
                'popup': "3 rue des coquelicots<br>01000 Libreville",
            }
        ]


@pytest.mark.django_db
class TestFormationAdminCreate:
    def setup_method(self):
        category = FormationCategoryFactory()
        self.url = reverse('formations_formation_modeladmin_create')
        self.valid_data = nested_form_data(
            {
                'title': "bling",
                'description': rich_text('plop'),
                'goal': "bling, blang",
                'category': category.id,
                'max_attendees': 0,
                'more_description': inline_formset([]),
                'pictures': inline_formset([]),
                'ranges': inline_formset(
                    [
                        {'begin': '2019-10-08 18:15', 'end': '20:15'},
                        {'begin': '2019-11-08 18:15', 'end': '20:15'},
                    ]
                ),
                'location': "Cliss XXI",
                'rue': "23, rue Jean Jaurès",
                'code_postal': "62510",
                'ville': "Liévin",
            }
        )

    def test_get_anonymous(self, client):
        response = client.get(self.url)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse('wagtailadmin_login'), self.url
        )

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_get(self, client, logged_user):
        assert log_in_client(
            client,
            logged_user(
                permissions__codenames=[
                    'add_formation',
                    'subscribe_formation',
                    'access_admin',
                ]
            ),
        )

        response = client.get(self.url)
        assert response.status_code == 200

    def test_post_anonymous(self, client):
        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse('wagtailadmin_login'), self.url
        )

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post(self, client, mailoutbox, logged_user):
        account = logged_user(
            permissions__codenames=[
                'add_formation',
                'subscribe_formation',
                'access_admin',
            ]
        )
        assert log_in_client(client, account)

        response = client.post(self.url, self.valid_data)
        messages = list(get_messages(response.wsgi_request))
        assert response.status_code == 302
        assert response['Location'] == reverse(
            'formations_formation_modeladmin_index'
        )

        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

        formation = account.formations.last()
        assert formation.ranges.count() == 2
        begin_at = datetime.datetime.fromisoformat('2019-10-08 18:15')
        end_at = datetime.datetime.fromisoformat('2019-11-08 20:15')
        assert timezone.localtime(formation.begin_at) == timezone.make_aware(
            begin_at
        )
        assert timezone.localtime(formation.end_at) == timezone.make_aware(
            end_at
        )


@pytest.mark.django_db
class TestFormationSubscribe:
    def setup_method(self):
        self.formation = FormationFactory()
        self.url = reverse(
            'formations:formation-subscribe', args=[self.formation.pk]
        )
        self.valid_data = {
            'name': "bling",
            'family_name': "blang",
            'email': "bling@blang.net",
            'phone': "01 24 54 87 98",
            'rue': "42, rue Jean Jaurès",
            'code_postal': "62510",
            'ville': "Liévin",
        }

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_get(self, client, logged_user):
        assert log_in_client(client, logged_user())

        response = client.get(self.url)
        assert response.status_code == 200

    def test_post_anonymous(self, client):
        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse(settings.LOGIN_URL), self.url
        )

    def test_post_owner(self, client, mailoutbox):
        assert log_in_client(client, self.formation.owner)

        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == self.formation.get_absolute_url()

        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

        # un courriel est envoyé à l'organisateur
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [self.formation.owner.email]

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other_invalid_0(self, client, mailoutbox, logged_user):
        assert log_in_client(client, logged_user())

        response = client.post(self.url)

        # les données sont requises
        assert response.status_code == 200
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == ERROR_LEVEL

        assert len(mailoutbox) == 0

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other_invalid_1(self, client, mailoutbox, logged_user):
        assert log_in_client(client, logged_user())

        response = client.post(
            self.url, {k: v for k, v in self.valid_data.items() if 'name' != k}
        )

        # une erreur
        assert response.status_code == 200
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == ERROR_LEVEL

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other_invalid_2(self, client, mailoutbox, logged_user):
        assert log_in_client(client, logged_user())

        response = client.post(self.url, {'name': 'bling', 'email': 'blang'})

        # plusieurs erreurs
        assert response.status_code == 200
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == ERROR_LEVEL
        assert '<br>' in messages[0].message

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_other(self, client, mailoutbox, logged_user):
        account = logged_user()
        assert log_in_client(client, account)

        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == self.formation.get_absolute_url()

        # un message de réussite est affiché
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

        # un courriel est envoyé à l'organisateur
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [self.formation.owner.email]

    @pytest.mark.parametrize(
        'logged_user',
        (AssociationUserAccountFactory, PersonneUserAccountFactory),
    )
    def test_post_lacking_space(self, client, mailoutbox, logged_user):
        self.formation.max_attendees = 1
        self.formation.save()
        self.formation.subscriptions.create(**self.valid_data, is_active=True)
        account = logged_user()
        assert log_in_client(client, account)

        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == self.formation.get_absolute_url()

        # un message d'avertissement est affiché
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == WARNING_LEVEL

        # un courriel est envoyé à l'organisateur
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [self.formation.owner.email]


@pytest.mark.django_db
class TestSubscriptionModerationAdminList:
    url = reverse_lazy('wagtailadmin_home')

    def test_get(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        formation = FormationFactory(owner=admin, is_active=True)
        subscriptions = SubscriptionFactory.create_batch(
            4, formation=formation, is_active=False
        )

        assert log_in_client(client, admin)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['subscriptions_moderation']) == len(
            subscriptions
        )


@pytest.mark.django_db
class TestSubscriptionModeration:
    def setup_method(self):
        self.admin = AssociationUserAccountFactory(is_superuser=True)
        self.formation = FormationFactory(owner=self.admin, is_active=True)
        self.subscriptions = SubscriptionFactory.create_batch(
            4, formation=self.formation, is_active=False
        )

    def test_approve(self, client, mailoutbox):
        assert log_in_client(client, self.admin)
        subscription = self.subscriptions[0]
        assert not subscription.is_active
        response = client.post(
            reverse('formations:approve-subscription', args=[subscription.id])
        )
        assert response.status_code == 302
        subscription.refresh_from_db()
        assert subscription.is_active

        # un courriel est envoyé à l'inscrit
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [subscription.email]

        # un message de succes est affiché au modérateur
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

    def test_approve_via_direct_edition(self, client, mailoutbox):
        assert log_in_client(client, self.admin)
        subscription = self.subscriptions[0]
        assert not subscription.is_active
        url = reverse(
            'formations_subscription_modeladmin_edit', args=[subscription.id]
        )
        response = client.get(url)
        assert response.status_code == 200

        # on copie le formulaire dans data
        form = (
            bs(response.content, 'html.parser')
            .find('div', class_='content')
            .find('form', attrs={'method': 'POST'})
        )
        data = {
            i.get('name'): i.get('value')
            for i in form.find_all('input')
            if i.get('value')
        }

        data['is_active'] = True
        response = client.post(url, data)
        assert response.status_code == 302
        subscription.refresh_from_db()
        assert subscription.is_active

        # un courriel est envoyé à l'inscrit
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [subscription.email]

        # deux messages de succes sont affiché au modérateur
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 2
        assert messages[0].level == SUCCESS_LEVEL
        assert messages[1].level == SUCCESS_LEVEL

    def test_reject(self, client, mailoutbox):
        assert log_in_client(client, self.admin)
        subscription_id = self.subscriptions[0].id
        subscription_email = self.subscriptions[0].email
        response = client.post(
            reverse('formations:reject-subscription', args=[subscription_id])
        )
        assert response.status_code == 302
        assert not self.formation.subscriptions.filter(id=subscription_id)

        # un courriel est envoyé à l'inscrit
        assert len(mailoutbox) == 1
        assert mailoutbox[0].to == [subscription_email]

        # un message de warning est affiché au modérateur
        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == WARNING_LEVEL


@pytest.mark.django_db
class TestFormationAdminList:
    url = reverse_lazy('formations_formation_modeladmin_index')

    def test_get(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        formations = FormationFactory.create_batch(4, owner=admin)

        assert log_in_client(client, admin)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['object_list']) == len(formations)


@pytest.mark.django_db
class TestFormationAdminEdit:
    def test_get(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        formation = FormationFactory(owner=admin)
        SubscriptionFactory.create_batch(4, formation=formation)

        assert log_in_client(client, admin)

        url = reverse_lazy(
            'formations_formation_modeladmin_edit', args=[formation.id]
        )
        response = client.get(url)
        assert response.status_code == 200


@pytest.mark.django_db
class TestSubscriptionAdminList:
    url = reverse_lazy('formations_subscription_modeladmin_index')

    def test_get(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        formation = FormationFactory(owner=admin)
        subscriptions = SubscriptionFactory.create_batch(4, formation=formation)

        assert log_in_client(client, admin)

        response = client.get(self.url)
        assert response.status_code == 200
        assert len(response.context['object_list']) == len(subscriptions)


@pytest.mark.django_db
class TestSubscriptionAdminEdit:
    def setup_method(self):
        self.formation = FormationFactory()
        SubscriptionFactory(formation=self.formation)
        self.valid_data = {
            'formation': self.formation.id,
            'name': "bling",
            'family_name': "blang",
            'email': "bling@blang.net",
            'phone': "01 24 54 87 98",
            'rue': "42, rue Jean Jaurès",
            'code_postal': "62510",
            'ville': "Liévin",
        }
        self.url = reverse(
            'formations_subscription_modeladmin_edit',
            args=(self.formation.subscriptions.last().id,),
        )

    def test_get_anonymous(self, client):
        response = client.get(self.url)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse('wagtailadmin_login'), self.url
        )

    def test_post_anonymous(self, client):
        response = client.post(self.url, self.valid_data)
        assert response.status_code == 302
        assert response['Location'] == '{}?next={}'.format(
            reverse('wagtailadmin_login'), self.url
        )

    def test_get(self, client):
        assert log_in_client(client, self.formation.owner)

        response = client.get(self.url)
        assert response.status_code == 200

    def test_post(self, client, mailoutbox):
        assert log_in_client(client, self.formation.owner)

        response = client.post(self.url, self.valid_data)

        assert response.status_code == 302
        assert response['Location'] == "{}#tab-inscriptions".format(
            reverse(
                'formations_formation_modeladmin_edit',
                args=(self.formation.id,),
            )
        )

        messages = list(get_messages(response.wsgi_request))
        assert len(messages) == 1
        assert messages[0].level == SUCCESS_LEVEL

        # finally, fetching the success url should be ok
        response = client.get(response['Location'])
        assert response.status_code == 200


@pytest.mark.django_db
class TestHelpView:
    def test_get(self, client):
        url = reverse_lazy('formations:help')
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_get_edit_form(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse_lazy(
            'wagtailsettings:edit', args=['formations', 'formationssettings']
        )
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_follow_category_button(self, client):
        admin = AssociationUserAccountFactory(is_superuser=True)
        assert log_in_client(client, admin)

        url = reverse_lazy(
            'wagtailsettings:edit', args=['formations', 'formationssettings']
        )
        response = client.get(url, follow=True)
        assert response.status_code == 200
        buttons = bs(response.content, 'html.parser').find_all(
            'a', class_='button'
        )
        assert len(buttons) > 0
        url = buttons[0].attrs['href']
        response = client.get(url)
        assert response.status_code == 200


@pytest.mark.django_db
class TestHomeSummaryItems:
    def test_get_can_add(self, client):
        assert log_in_client(
            client,
            AssociationUserAccountFactory(
                permissions__codenames=['add_formation', 'access_admin']
            ),
        )
        url = reverse_lazy('wagtailadmin_home')
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_get_can_edit(self, client):
        assert log_in_client(
            client,
            AssociationUserAccountFactory(
                permissions__codenames=['edit_formation', 'access_admin']
            ),
        )
        url = reverse_lazy('wagtailadmin_home')
        response = client.get(url, follow=True)
        assert response.status_code == 200

    def test_get(self, client):
        assert log_in_client(
            client,
            AssociationUserAccountFactory(
                permissions__codenames=['access_admin']
            ),
        )
        url = reverse_lazy('wagtailadmin_home')
        response = client.get(url, follow=True)
        assert response.status_code == 200
