from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import AccessMixin
from django.forms import Form
from django.http import Http404
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.views.generic import DetailView, FormView, ListView, TemplateView
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import BaseFormView, FormMixin

from wagtail.contrib.modeladmin import views as modeladmin_views
from wagtail.models import Site

from boiteaasso.base.mixins import ConfirmationEmailMixin, MapViewMixin
from boiteaasso.utils.views import CollectionPageMixin

from .forms import FormationsFiltersForm, FormationSubscribeForm
from .models import (
    Formation,
    FormationCategory,
    FormationsSettings,
    Subscription,
)


class SettingsMixin:
    @cached_property
    def settings(self):
        """Retourne les paramètres du service de formations."""
        return FormationsSettings.for_site(Site.find_for_request(self.request))

    @property
    def help_url(self):
        return (
            None
            if not self.settings.help_is_active
            else self.request.build_absolute_uri(reverse('formations:help'))
        )


class FormationsPageMixin(CollectionPageMixin, SettingsMixin):
    collection_list_title = "Formations"
    collection_list_url = reverse_lazy('formations:formations')
    show_breadcrumb = False

    def collection_include_list_crumb(self):
        return issubclass(self.__class__, DetailView)

    def get_breadcrumb(self):
        return [
            {'title': "Services"},
            {'title': "Formations"},
        ] + super().get_breadcrumb()


class HelpView(FormationsPageMixin, TemplateView):
    template_name = 'formations/help.html'
    page_title = "Aide pour les formations"
    breadcrumb_title = "Aide"

    def get(self, request, *args, **kwargs):
        if not self.settings.help_is_active:
            return redirect(self.collection_list_url)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['content'] = self.settings.help_content
        return context

    def get_page_actions(self):
        return [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les formations",
                'label': self.collection_list_title,
            }
        ]


class FormationListView(FormationsPageMixin, MapViewMixin, ListView):
    """
    Liste les formations, éventuellement filtrées.
    """

    context_object_name = 'formations'
    template_name = 'formations/formation_list.html'
    page_seo_title = "Liste des formations"
    page_title = "Formations"

    def get_queryset(self):
        """Filtre la sélection de formations à afficher."""
        queryset = Formation.objects.active().future()
        # filtre avec la categorie définis dans l'URL
        self.formations_category = self.kwargs.get('formations_category', None)
        if self.formations_category is not None:
            try:
                self.formations_category = FormationCategory.objects.get(
                    slug=self.formations_category
                )
            except Exception:
                raise Http404(
                    "La categorie de formation demandée est inconnue."
                )
            queryset = queryset.filter(category=self.formations_category)
        # filtre avec les autres critères du formulaire
        self.filters_form = FormationsFiltersForm(self.request.GET)
        if self.filters_form.has_filters:
            self.filters_form.full_clean()
            queryset = self.filters_form.filter_queryset(queryset)
        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['formations_category'] = self.formations_category
        context['filters_form'] = self.filters_form
        context['introduction'] = self.settings.introduction
        return context

    def get_page_actions(self):
        actions = []
        if self.request.user.has_perm('formations.add_formation'):
            actions.append(
                {
                    'url': '{}?{}'.format(
                        reverse('formations_formation_modeladmin_index'),
                        FormationIndexView.USER_FORMATIONS_VAR,
                    ),
                    'class': 'icon fa-graduation-cap',
                    'title': "Gérer mes formations",
                    'label': "Mes formations",
                }
            )
        if self.settings.help_is_active:
            actions.append(
                {
                    'url': reverse('formations:help'),
                    'title': HelpView.page_title,
                    'label': mark_safe(
                        '<i class="fa fa-question-circle" aria-hidden="true">'
                        '</i>'
                    ),
                    'attrs': {'aria-label': "Aide sur les formations"},
                }
            )
        return actions

    def get_map_options(self, context):
        return {
            'markers': self.get_formations_markers(context['object_list']),
            'mapOptions': {'center': settings.DEFAULT_MAP_CENTER, 'zoom': 8},
        }

    def get_formations_markers(self, object_list):
        return [
            {
                'lat': formation.latitude,
                'lng': formation.longitude,
                'popup': render_to_string(
                    'formations/formation_popup.html', {'formation': formation}
                ),
            }
            for formation in object_list
            if formation.is_geolocated
        ]


class SubscribeFormMixin(FormMixin):
    form_class = FormationSubscribeForm

    def get_initial(self):
        initial = super().get_initial()
        if not self.request.user.is_anonymous:
            initial.update(
                {
                    'name': self.request.user.short_name,
                    'family_name': self.request.user.full_name,
                    'email': self.request.user.email,
                }
            )
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['formation'] = self.object
        return kwargs


class FormationDetailView(
    FormationsPageMixin, MapViewMixin, SubscribeFormMixin, DetailView
):
    """
    Présente une formation.
    """

    context_object_name = 'formation'
    queryset = Formation.objects.active()
    template_name = 'formations/formation.html'

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)
        # récupère les permissions de l'utilisateur sur la formation
        self.user_perms = obj.permissions_for_user(self.request.user)
        return obj

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['can_subscribe'] = self.user_perms.can_subscribe()
        context['form'] = self.get_form()
        return context

    def get_page_title(self):
        return self.object.title

    def get_page_seo_title(self):
        return '{} : {}'.format(self.object.category, self.object.title)

    def get_page_actions(self):
        actions = [
            {
                'url': self.collection_list_url,
                'class': 'icon fa-list',
                'title': "Voir les formations",
                'label': self.collection_list_title,
            }
        ]
        if self.user_perms.can_edit():
            actions += [
                {
                    'url': reverse(
                        'formations_formation_modeladmin_edit',
                        args=[self.object.pk],
                    ),
                    'class': 'icon fa-pencil',
                    'title': "Modifier cette formation",
                    'label': "Modifier…",
                },
                {
                    'url': "{}#tab-inscriptions".format(
                        reverse(
                            'formations_formation_modeladmin_edit',
                            args=[self.object.pk],
                        )
                    ),
                    'class': 'icon fa-ticket',
                    'title': "Gérer les inscriptions à cette formation",
                    'label': "Inscriptions…",
                },
            ]
        return actions

    def get_map_options(self, context):
        if not self.object.is_geolocated:
            return {}
        return {
            'markers': [
                {
                    'lat': self.object.latitude,
                    'lng': self.object.longitude,
                    'popup': self.object.get_adresse_display(),
                }
            ],
            'mapOptions': {
                'center': [self.object.latitude, self.object.longitude],
                'zoom': 8,
            },
        }


# VUES RESTREINTES
# ------------------------------------------------------------------------------


class FormationSubscribeView(
    AccessMixin,
    ConfirmationEmailMixin,
    FormationDetailView,
    BaseFormView,
):
    """
    Inscrit l'utilisateur à une formation ; notifie l'organisateur-trice.
    """

    template_name = 'formations/formation_subscribe.html'
    email_template = 'formations/emails/formation_subscription.txt'
    email_template_html = 'formations/emails/formation_subscription.html'
    email_template_subject = (
        'formations/emails/formation_subscription_subject.txt'
    )

    def has_permission(self):
        return self.object.permissions_for_user(
            self.request.user
        ).can_subscribe()

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_permission():
            if not request.user.is_authenticated:
                return self.handle_no_permission()
            messages.error(
                self.request,
                "Vous n'avez pas la permission de vous inscrire à cette "
                "formation.",
            )
            return redirect(self.object)
        return super().dispatch(request, *args, **kwargs)

    # Traitement du formulaire ------------------------------------------------

    def process_form_subscription(self, form):
        subscription = form.save(commit=False)
        subscription.formation = self.object
        subscription.user = self.request.user
        subscription.save()
        return subscription

    def form_valid(self, form):
        self.process_form_subscription(form)
        self.send_confirmation_email(self.object.owner)

        message = (
            "Votre demande d'inscription pour la formation « {} » a été "
            "prise en compte. L'organisateur en a été informé.".format(
                self.object.title
            )
        )
        if self.object.is_lacking_places():
            message += (
                " Ceci étant, la formation risque de manquer de place pour "
                "accueillir tous les participants. L'organisateur vous "
                "informera ultérieurement si votre inscription est confirmée "
                "ou si malheureusement il ne peut y donner suite."
            )
            messages.warning(self.request, message)
        else:
            messages.success(self.request, message)

        return redirect(self.object)

    def form_invalid(self, form):
        errors = (
            set().union(*form.errors.values()).union(form.non_field_errors())
        )
        if not errors:
            message = "Impossible de vous inscrire à la formation."
        elif len(errors) == 1:
            message = "Impossible de vous inscrire à la formation : {}".format(
                errors.pop().lower()
            )
        else:
            message = "Impossible de vous inscrire à la formation :{}".format(
                '<br>&nbsp;- '.join(['', *errors])
            )
        messages.error(self.request, message)

        return super().form_invalid(form)


class FireModelSaveView:
    """
    ClusterForm save's method save related in such a manner it is
    impossible to see related in model's save method. So we need to refire
    it during form handling.
    """

    def form_valid(self, form):
        r = super().form_valid(form)
        form.instance.save()
        return r


class FormationEditView(FireModelSaveView, modeladmin_views.EditView):
    pass


class FormationCreateView(FireModelSaveView, modeladmin_views.CreateView):
    def get_instance(self):
        # définis l'organisateur-trice avec la personne connectée
        instance = super().get_instance()
        instance.owner = self.request.user
        return instance


class FormationIndexView(modeladmin_views.IndexView):
    USER_FORMATIONS_VAR = 'self'
    # ajoute un argument pour filtrer avec les formations de la personne
    IGNORED_PARAMS = modeladmin_views.IndexView.IGNORED_PARAMS + (
        USER_FORMATIONS_VAR,
    )


class SubscriptionEditView(ConfirmationEmailMixin, modeladmin_views.EditView):
    email_template = 'formations/emails/subscription_approved.txt'
    email_template_html = 'formations/emails/subscription_approved.html'
    email_template_subject = (
        'formations/emails/subscription_approved_subject.txt'
    )

    def post(self, *args, **kwargs):
        # We store active before form processing, since the framework weirdly
        # modify self.instance when form.is_valid is called in ProcessFormView.
        self.was_active = self.instance.is_active
        return super().post(*args, **kwargs)

    def form_valid(self, form):
        if form.cleaned_data['is_active'] and not self.was_active:
            self.send_confirmation_email(FakeUser(**form.cleaned_data))
            m = "Inscription confirmée. Un courriel a été expédié à l'inscrit."
            messages.success(self.request, m)
        if not form.cleaned_data['is_active'] and self.was_active:
            m = (
                "Vous avez annulé manuellement la confirmation d'une "
                "inscription dont la confirmation avait été expédiée "
                "à l'inscrit. Prenez soin svp d'informer la personne."
            )
            messages.warning(self.request, m)
        return super().form_valid(form)

    def get_success_url(self):
        # renvoit sur une page accessible à l'organisateur
        return "{}#tab-inscriptions".format(
            reverse(
                'formations_formation_modeladmin_edit',
                args=(self.instance.formation_id,),
            )
        )


class SubscriptionIndexView(modeladmin_views.IndexView):
    USER_SUBSCRIPTIONS_VAR = 'self'
    # ajoute un argument pour filtrer avec les subscriptions de la personne
    IGNORED_PARAMS = modeladmin_views.IndexView.IGNORED_PARAMS + (
        USER_SUBSCRIPTIONS_VAR,
    )


# MODERATION DES INSCRIPTIONS
# ------------------------------------------------------------------------------


class SubscriptionModerationMixin(
    AccessMixin,
    SingleObjectMixin,
    SettingsMixin,
    FormView,
    ConfirmationEmailMixin,
):
    model = Subscription
    form_class = Form
    success_url = reverse_lazy('wagtailadmin_home')

    def has_permission(self):
        return (
            self.request.user.is_superuser
            or self.object.formation.owner == self.request.user
        )

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_permission():
            if not request.user.is_authenticated:
                return self.handle_no_permission()
            messages.error(
                self.request, "Vous n'avez pas la permission pour cette action"
            )
            return redirect('wagtailadmin_home')
        return super().dispatch(request, *args, **kwargs)


class FakeUser:
    """Émule un objet User à partir des données de souscription."""

    def __init__(self, name, family_name, email, **kwargs):
        self.email = email
        self.name = name
        self.family_name = family_name

    def __str__(self):
        return "{} {}".format(self.name, self.family_name)


class SubscriptionAppproveView(SubscriptionModerationMixin):
    form_class = Form
    success_url = reverse_lazy('wagtailadmin_home')

    email_template = 'formations/emails/subscription_approved.txt'
    email_template_html = 'formations/emails/subscription_approved.html'
    email_template_subject = (
        'formations/emails/subscription_approved_subject.txt'
    )

    def form_valid(self, form):
        self.object.is_active = True
        self.object.save()
        self.send_confirmation_email(FakeUser(**self.object.__dict__))
        m = "Inscription confirmée. Un courriel a été expédié à l'inscrit."
        messages.success(self.request, m)
        return super().form_valid(form)


class SubscriptionRejectView(SubscriptionModerationMixin):
    form_class = Form
    success_url = reverse_lazy('wagtailadmin_home')

    email_template = 'formations/emails/subscription_rejected.txt'
    email_template_html = 'formations/emails/subscription_rejected.html'
    email_template_subject = (
        'formations/emails/subscription_rejected_subject.txt'
    )

    def form_valid(self, form):
        self.send_confirmation_email(FakeUser(**self.object.__dict__))
        m = "Inscription rejetée. Un courriel a été expédié à l'inscrit."
        messages.warning(self.request, m)
        self.object.delete()
        return super().form_valid(form)
