from django.contrib.admin import SimpleListFilter
from django.db.models import Q
from django.http.response import HttpResponseRedirect
from django.template.loader import render_to_string
from django.urls import re_path
from django.utils.html import format_html
from django.utils.translation import gettext as _

from wagtail import hooks
from wagtail.admin.panels import Panel
from wagtail.admin.site_summary import SummaryItem
from wagtail.contrib.modeladmin.helpers import (
    AdminURLHelper,
    ButtonHelper,
    PermissionHelper,
)
from wagtail.contrib.modeladmin.options import ModelAdmin

from boiteaasso.base.templatetags.horodated import postdatetime

from .models import (
    Formation,
    FormationDescription,
    FormationPicture,
    FormationTimeRange,
    Subscription,
)
from .views import (
    FormationCreateView,
    FormationEditView,
    FormationIndexView,
    SubscriptionEditView,
    SubscriptionIndexView,
)

# TABLEAU DE BORD
# ------------------------------------------------------------------------------


class FormationsSummaryItem(SummaryItem):
    """
    Affiche le nombre de formations de la personne ou au total.
    """

    order = 910
    template_name = 'formations/dashboard/summary_formations.html'

    def __init__(self, request, user_only):
        super().__init__(request)
        self.user_only = user_only

    def get_context_data(self, parent_context):
        formations = Formation.objects.all()
        if self.user_only:
            formations = formations.filter(owner=self.request.user)
        return {'total_formations': formations.count()}


@hooks.register('construct_homepage_summary_items')
def add_formations_summary_item(request, items):
    """Ajoute le nombre d'inscription sur le tableau de bord."""
    user_only = None
    if request.user.has_perm('formations.edit_formation'):
        user_only = False
    elif request.user.has_perm('formations.add_formation'):
        user_only = True
    if user_only is not None:
        items.append(FormationsSummaryItem(request, user_only))


class FormationsSubscriptionsSummaryItem(SummaryItem):
    """
    Ajoute le nombre de formations sur le tableau de bord.
    """

    order = 911
    template_name = 'formations/dashboard/summary_subscriptions.html'

    def __init__(self, request, manageable):
        super().__init__(request)
        self.manageable = manageable

    def get_context_data(self, parent_context):
        subscriptions = Subscription.objects.filter(user=self.request.user)
        return {
            'total_subscriptions': subscriptions.count(),
            'manageable': self.manageable,
        }


@hooks.register('construct_homepage_summary_items')
def add_subscripions_summary_item(request, items):
    """
    Affiche le nombre d'inscription de la personne ou au total.
    """
    manage = None
    if request.user.has_perm(
        'formations.edit_subscription'
    ) or request.user.has_perm('formations.add_subscription'):
        manage = True
    items.append(FormationsSubscriptionsSummaryItem(request, manage))


class SubscriptionsForModerationPanel(Panel.BoundPanel):
    name = 'subscriptions_for_moderation'
    order = 201

    def __init__(self, request):
        self.request = request
        if request.user.is_superuser:
            self.subscriptions_moderation = Subscription.objects.filter(
                is_active=False
            )
        else:
            self.subscriptions_moderation = Subscription.objects.filter(
                is_active=False, formation__owner=request.user
            )

    def render_html(self, parent_context):
        return render_to_string(
            'formations/dashboard/subscriptions_for_moderation.html',
            {'subscriptions_moderation': self.subscriptions_moderation},
            request=self.request,
        )


@hooks.register('construct_homepage_panels')
def add_subscriptions_moderation_panel(request, panels):
    panels.append(SubscriptionsForModerationPanel(request))


# GESTION DES FORMATIONS
# ------------------------------------------------------------------------------


class FormationButtonHelper(ButtonHelper):
    def add_button(self, *args, **kwargs):
        # améliore le texte du bouton d'ajout
        button = super().add_button(*args, **kwargs)
        button['label'] = "Ajouter une formation"
        button['title'] = "Ajouter une nouvelle formation"
        return button

    def get_buttons_for_obj(self, formation, **kwargs):
        buttons = super().get_buttons_for_obj(formation, **kwargs)
        url_helper = AdminURLHelper(self.model)

        def button(action_url, label, classnames):
            return {
                'url': url_helper.get_action_url(
                    action_url, instance_pk=formation.id
                ),
                'label': label,
                'title': label,
                'classname': 'button button-small ' + classnames,
            }

        copy_btn = button('copy', _('Copier'), 'button-secondary')
        buttons.append(copy_btn)
        return buttons


class FormationPermissionHelper(PermissionHelper):
    def user_can_list(self, user):
        # affiche la liste et l'élément de menu si l'utilisateur
        # peut ajouter une formation ou les modifier
        return self.user_can_create(user) or self.user_can_edit(user)

    def user_can_edit(self, user):
        perm_codename = self.get_perm_codename('change')
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_inspect_obj(self, user, obj):
        if not self.inspect_view_enabled:
            return False
        perms = obj.permissions_for_user(user)
        return perms.is_owner()

    def user_can_edit_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_edit()

    def user_can_delete_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_delete()


class OldFormationListFilter(SimpleListFilter):
    title = 'ancienneté'
    parameter_name = 'is_past'

    def lookups(self, request, model_admin):
        return (('yes', 'Formations anciennes'),)

    def queryset(self, request, queryset):
        if self.value() == 'yes':
            return queryset
        else:
            return queryset.recent()

    def choices(self, changelist):
        yield {
            'selected': self.value() is None,
            'query_string': changelist.get_query_string(
                remove=[self.parameter_name]
            ),
            'display': 'Formations récentes',
        }
        for lookup, title in self.lookup_choices:
            yield {
                'selected': self.value() == str(lookup),
                'query_string': changelist.get_query_string(
                    {self.parameter_name: lookup}
                ),
                'display': title,
            }


class FormationAdmin(ModelAdmin):
    """
    Administration des formations pour la personne connectée.
    """

    model = Formation
    menu_label = "Formations"
    menu_icon = 'graduation-cap'
    # étends les classes utilisées par défaut
    button_helper_class = FormationButtonHelper
    permission_helper_class = FormationPermissionHelper
    edit_view_class = FormationEditView
    create_view_class = FormationCreateView
    index_view_class = FormationIndexView
    # personnalise la liste des formations
    ordering = ['-last_modified_at']
    list_display = (
        'picture',
        'title_formatted',
        'start_when_formatted',
        'category',
        'updated',
        'is_active',
    )
    list_display_add_buttons = 'title_formatted'
    list_filter = ('category', 'is_active', OldFormationListFilter)
    list_per_page = 15

    def get_queryset(self, request):
        qs = super().get_queryset(request).select_related('owner')
        if self.is_user_formations_only(request):
            # filtre avec les formations de la personne
            qs = qs.filter(owner=request.user)
        return qs

    def get_list_display(self, request):
        list_display = self.list_display
        if not self.is_user_formations_only(request):
            # affiche également l'organisateur-trice
            list_display += ('owner_formatted',)
        return list_display

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'title_formatted':
            return ['title']
        return []

    def get_extra_attrs_for_row(self, obj, context):
        if not obj.is_active:
            return {'class': 'unpublished'}
        return {}

    def is_user_formations_only(self, request):
        """Détermine s'il ne faut afficher que les formations de la personne."""
        return (
            FormationIndexView.USER_FORMATIONS_VAR in request.GET
            or not request.user.has_perm('formations.change_formation')
        )

    def get_admin_urls_for_registration(self):
        urls = super().get_admin_urls_for_registration()

        def gen_url(pattern, view, name=None):
            if not name:
                name = pattern
            return re_path(
                self.url_helper.get_action_url_pattern(pattern),
                view,
                name=self.url_helper.get_action_url_name(name),
            )

        urls += (gen_url('copy', self.copy),)
        return urls

    def copy(self, request, instance_pk):
        instance = self.model.objects.get(pk=instance_pk)
        instance.title = '{} ({})'.format(instance.title, _('Copie'))
        instance.pk = None
        instance.id = None
        instance.is_active = False
        instance.save()
        for fdescription in FormationDescription.objects.filter(
            formation=instance_pk
        ):
            copy = fdescription.copy(formation=instance)
            copy.save()
        for ftimerange in FormationTimeRange.objects.filter(
            formation=instance_pk
        ):
            copy = ftimerange.copy(formation=instance)
            copy.save()
        for fpicture in FormationPicture.objects.filter(formation=instance_pk):
            copy = fpicture.copy(formation=instance)
            copy.save()
        return HttpResponseRedirect(self.url_helper.get_action_url('index'))

    # Mise en forme des champs ------------------------------------------------

    def picture(self, obj):
        picture = obj.get_highlighted_picture()
        if not picture:
            return ''
        return format_html(
            '<img src="{}" class="photo-thumb" alt="{}">',
            picture.picture.thumbnail.url,
            picture.title,
        )

    picture.short_description = "image"

    def start_when_formatted(self, obj):
        start = obj.ranges.first()
        return postdatetime(start.begin) if start else None

    start_when_formatted.short_description = "Démarre le"

    def title_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette formation">{}</a></h2>',
            obj.get_absolute_url(),
            obj.title,
        )

    title_formatted.short_description = "titre"
    title_formatted.admin_order_field = 'title'

    def updated(self, obj):
        return postdatetime(obj.last_modified_at)

    updated.short_description = "mise à jour"
    updated.admin_order_field = 'last_modified_at'

    def owner_formatted(self, obj):
        return format_html(
            '<a href="{}" title="Voir les détails du compte">{}</a>',
            obj.owner.get_admin_details_url(),
            obj.owner.get_short_name() or obj.owner.email,
        )

    owner_formatted.short_description = "organisateur-trice"
    owner_formatted.admin_order_field = 'owner'


# GESTION DES INSCRIPTIONS
# ------------------------------------------------------------------------------


class SubscriptionButtonHelper(ButtonHelper):
    def add_button(self, *args, **kwargs):
        # supprime le bouton d'ajout
        return None


class SubscriptionPermissionHelper(PermissionHelper):
    def user_can_list(self, user):
        # affiche la liste et l'élément de menu si l'utilisateur
        # peut ajouter une subscription ou les modifier
        return self.user_can_create(user) or self.user_can_edit(user)

    def user_can_edit(self, user):
        perm_codename = self.get_perm_codename('change')
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_inspect_obj(self, user, obj):
        if not self.inspect_view_enabled:
            return False
        perms = obj.permissions_for_user(user)
        return perms.is_owner()

    def user_can_edit_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_edit()

    def user_can_delete_obj(self, user, obj):
        perms = obj.permissions_for_user(user)
        return perms.can_delete()


class SubscriptionAdmin(ModelAdmin):
    model = Subscription
    is_pagemodel = False
    permission_helper = None
    url_helper = None
    menu_label = "Inscriptions à des formations"
    menu_icon = 'ticket'
    # étends les classes utilisées par défaut
    button_helper_class = SubscriptionButtonHelper
    permission_helper_class = SubscriptionPermissionHelper
    create_view_class = None
    edit_view_class = SubscriptionEditView
    index_view_class = SubscriptionIndexView
    # personnalise la liste des inscriptions
    ordering = ['-submit_time']
    list_display = (
        'formation_formatted',
        'is_active',
        'subscribed',
        'name',
        'family_name',
        'email',
    )
    list_display_add_buttons = 'is_active'
    list_filter = ('formation', 'is_active', OldFormationListFilter)
    list_per_page = 15

    def has_add_permission(self, request):
        return False

    def is_priviledged_user(self, request):
        return request.user.is_superuser or request.user.has_perm(
            'formations.change_subscription'
        )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        if self.is_priviledged_user(request):
            return qs
        return qs.filter(Q(user=request.user) | Q(formation_owner=request.user))

    def get_list_display(self, request):
        list_display = self.list_display
        if self.is_priviledged_user(request):
            # affiche également le compte d'inscription
            list_display += ('user_formatted',)
        return list_display

    def get_extra_attrs_for_row(self, obj, context):
        if not obj.is_active:
            return {'class': 'unpublished'}
        return {}

    def get_extra_class_names_for_field_col(self, obj, field_name):
        if field_name == 'formation_formatted':
            return ['title']
        return []

    # Mise en forme des champs ------------------------------------------------

    def formation_formatted(self, obj):
        return format_html(
            '<h2><a href="{}" title="Voir cette formation">{}</a></h2>',
            obj.formation.get_absolute_url(),
            obj.formation,
        )

    formation_formatted.short_description = "Formation"
    formation_formatted.admin_order_field = 'formation'

    def subscribed(self, obj):
        return postdatetime(obj.submit_time)

    subscribed.short_description = "inscrit le"
    subscribed.admin_order_field = 'submit_time'

    def user_formatted(self, obj):
        return format_html(
            '<a href="{}" title="Voir les détails du compte">{}</a>',
            obj.user.get_admin_details_url(),
            obj.user.get_short_name() or obj.user.email,
        )

    user_formatted.short_description = "compte d'inscription"
    user_formatted.admin_order_field = 'user'
