from django.contrib.admin.utils import label_for_field

from wagtail.contrib.modeladmin.views import InspectView


class RelatedInspectView(InspectView):
    """
    Étends la vue d'inspection d'un modèle afin de permettre d'afficher des
    champs d'un autre modèle lié. Ils sont à ajouter dans la propriété
    `inspect_view_fields` de la classe `ModelAdmin` pour ce modèle, en
    suivant le même format que pour traverser des relations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # récupère l'instance avec ses objets liés
        select_related = [
            field_name.rsplit('__', 1)[0]
            for field_name in self.model_admin.get_inspect_view_fields()
            if '__' in field_name
        ]
        if select_related:
            # FIXME: éviter la double requête en sélectionnant les
            # objets liés à la construction de la queryset, voir :
            # InstanceSpecificView.__init__()
            self.instance = self.model.objects.select_related(
                *select_related
            ).get(pk=self.instance_pk)

    def get_field_label(self, field_name, field=None):
        """Return a label to display for a field"""
        if hasattr(field, 'model'):
            return label_for_field(field_name, model=field.model)
        return label_for_field(field_name, model=self.model)

    def get_dict_for_field(self, field_name):
        if '__' not in field_name:
            return super().get_dict_for_field(field_name)

        related_names = field_name.split('__')
        field_name = related_names.pop(-1)

        # récupère le modèle de la relation et son instance
        related_instance = self.instance
        related_model = self.instance._meta.model
        for related_name in related_names:
            related_field = related_model._meta.get_field(related_name)
            related_model = related_field.related_model
            if related_instance is not None:
                related_instance = getattr(related_instance, related_name, None)

        # récupère le champ et sa valeur
        field = related_model._meta.get_field(field_name)
        if related_instance is not None:
            # FIXME: permettre de spécifier l'instance à utiliser, voir :
            # InspectView.get_field_display_value()
            orig_instance = self.instance
            self.instance = related_instance
            val = self.get_field_display_value(field_name, field)
            self.instance = orig_instance
        else:
            val = self.model_admin.get_empty_value_display(field_name)

        return {'label': self.get_field_label(field_name, field), 'value': val}
