from django.contrib.auth.models import AnonymousUser
from django.forms.widgets import Media
from django.views.generic import ListView, TemplateView

import pytest

from ..views import CollectionPageMixin, PageMixin


class Page(PageMixin, TemplateView):
    page_title = "Titre de la page"
    template_name = 'base.html'


class CollectionPage(CollectionPageMixin, TemplateView):
    page_title = "Titre de la page"
    template_name = 'base.html'
    collection_list_title = "Collection"
    collection_list_url = '/'


class TestPageMixin:
    @pytest.fixture(autouse=True)
    def setup_method(self, rf):
        self.request = rf.get('/')
        self.request.user = AnonymousUser()
        self.view = Page()
        self.view.request = self.request

    def test_page_title(self):
        assert self.view.get_page_title() == "Titre de la page"
        self.view.page_title = None
        with pytest.raises(ValueError):
            self.view.get_page_title()

    def test_page_seo_title(self):
        assert self.view.get_page_seo_title() is None
        self.view.page_seo_title = "Autre titre"
        assert self.view.get_page_seo_title() == "Autre titre"

    def test_page_media(self):
        assert isinstance(self.view.page_media, Media)

        class MediaPage(Page):
            class Media:
                css = {'all': ('css/foo.css',)}
                js = ('js/foo.js',)

        media = MediaPage().page_media
        assert 'css/foo.css' in media['css'].render()
        assert 'js/foo.js' in media['js'].render()

        class EmptyMediaPage(MediaPage):
            page_media = None

        assert EmptyMediaPage().page_media is None

    def test_page_actions(self):
        assert self.view.get_page_actions() == []

    def test_breadcrumb_title(self):
        assert self.view.get_breadcrumb_title() == "Titre de la page"
        self.view.breadcrumb_title = "Autre titre"
        assert self.view.get_breadcrumb_title() == "Autre titre"

    def test_breadcrumb(self):
        assert self.view.get_breadcrumb() == []

    def test_page_context(self):
        context = self.view.get_page_context()
        assert context.pop('title') == "Titre de la page"
        assert context.pop('seo_title') is None
        assert context.pop('actions') == []
        assert context.pop('breadcrumb') == [
            {'title': "Titre de la page", 'url': None}
        ]
        assert context.pop('media').render() == ''
        assert context.pop('show_breadcrumb') is True
        assert not context

    def test_context_data_page(self):
        assert self.view.get_context_data(page=None)['page'] is None

    def test_response(self):
        response = self.view.dispatch(self.request)
        assert response.status_code == 200
        assert 'page' in response.context_data


class TestCollectionPageMixin:
    @pytest.fixture(autouse=True)
    def setup_method(self, rf):
        self.request = rf.get('/')
        self.request.user = AnonymousUser()
        self.view = CollectionPage()
        self.view.request = self.request

    def test_collection_list_title(self):
        assert self.view.get_collection_list_title() == "Collection"
        self.view.collection_list_title = None
        with pytest.raises(ValueError):
            self.view.get_collection_list_title()

    def test_collection_list_url(self):
        assert self.view.get_collection_list_url() == '/'

    def test_breadcrumb(self):
        assert self.view.get_breadcrumb() == [
            {'title': "Collection", 'url': '/'}
        ]

    def test_breadcrumb_list_view(self):
        class List(CollectionPageMixin, ListView):
            collection_list_title = "Collection"

        assert List().get_breadcrumb() == []
