from django.forms.widgets import media_property
from django.views.generic import ListView

from wagtail.models import PAGE_TEMPLATE_VAR


class PageBase(type):
    """
    Métaclasse des vues présentant une page, qui ajoute une propriété
    'page_media' afin de pouvoir définir des médias depuis la page. Ils
    sont récupérés et construits depuis la classe `Meta` définie par la
    classe de la page, comme pour les formulaires de Django.
    """

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)

        if 'page_media' not in attrs:
            new_class.page_media = media_property(new_class)

        return new_class


class PageMixin(metaclass=PageBase):
    """
    Mixin à appliquer aux vues présentant une page. Il ajoute au contexte
    un dictionnaire `page`, sous la même forme qu'une Page de Wagtail,
    définissant principalement son titre.
    """

    #: Le titre de la page (requis).
    page_title = None

    #: Le titre de la page pour les moteurs de recherche.
    page_seo_title = None

    #: Une liste d'actions possibles en lien avec la page.
    page_actions = []

    #: Le titre de l'élément courant du le fil d'Ariane.
    breadcrumb_title = None

    #: Détermine s'il faut afficher le fil d'Ariane.
    show_breadcrumb = True

    def get_page_title(self):
        """Retourne le titre de la page."""
        if not self.page_title:
            raise ValueError(
                "La vue {} doit définir page_title.".format(
                    self.__class__.__name__
                )
            )
        return self.page_title

    def get_page_seo_title(self):
        """Retourne le titre de la page pour les moteurs de recherche."""
        return self.page_seo_title

    def get_page_actions(self):
        """Retourne la liste des actions possibles en lien avec la page.

        La liste retournée contient les boutons à afficher à côté du titre
        de la page. Ces boutons sont à définir sous forme de dictionnaire,
        pouvant contenir les attributs suivants :
            - url : l'URL vers lequel pointe le bouton ;
            - class : des classes supplémentaires à appliquer au bouton ;
            - title : un texte décrivant l'action du bouton ;
            - attrs : un dictionnaire d'attributs supplémentaires à définir
                      sur le bouton ;
            - label : le contenu du bouton.
        """
        return self.page_actions

    def get_breadcrumb_title(self):
        """
        Retourne le titre de l'élément courant du fil d'Ariane. Par défaut,
        le titre de la page est retourné.
        """
        return self.breadcrumb_title or self.get_page_title()

    def get_breadcrumb(self):
        """Retourne le fil d'Ariane sous forme d'une liste.

        Chaque élément du fil d'Ariane doit être un dictionnaire ou un
        objet avec au moins une clé ou propriété 'title'. Si 'url' est
        aussi définie, l'élément devient un lien actif.
        """
        return []

    def get_page_context(self):
        """Fournis le contexte à utiliser pour la page."""
        return {
            'title': self.get_page_title(),
            'seo_title': self.get_page_seo_title(),
            'actions': self.get_page_actions(),
            'breadcrumb': self.get_breadcrumb()
            + [{'title': self.get_breadcrumb_title(), 'url': None}],
            # propriétés de la page
            'media': self.page_media,
            'show_breadcrumb': self.show_breadcrumb,
        }

    def get_context_data(self, **kwargs):
        if PAGE_TEMPLATE_VAR not in kwargs:
            kwargs[PAGE_TEMPLATE_VAR] = self.get_page_context()
        return super().get_context_data(**kwargs)


class CollectionPageMixin(PageMixin):
    """
    Étends les fonctionnalités de `PageMixin` pour des vues liées à une
    collection de données. Le fil d'Ariane est notamment adapté pour y
    inclure un lien vers la page listant la collection.
    """

    #: Le titre de la page listant la collection (requis).
    collection_list_title = None

    #: L'URL de la page listant la collection.
    collection_list_url = None

    def get_collection_list_title(self):
        """Retourne le titre de la page listant la collection."""
        if not self.collection_list_title:
            raise ValueError(
                "La vue {} doit définir collection_list_title.".format(
                    self.__class__.__name__
                )
            )
        return self.collection_list_title

    def get_collection_list_url(self):
        """Retourne l'URL de la page listant la collection."""
        return self.collection_list_url

    def collection_include_list_crumb(self):
        """
        Détermine s'il faut inclure un élément pointant vers la page listant
        la collection dans le fil d'Ariane.
        """
        return not issubclass(self.__class__, ListView)

    def get_breadcrumb(self):
        breadcrumb = super().get_breadcrumb()

        if self.collection_include_list_crumb():
            breadcrumb.append(
                {
                    'title': self.get_collection_list_title(),
                    'url': self.get_collection_list_url(),
                }
            )

        return breadcrumb
